from collections import deque
from typing import Callable
from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import (
    Address,
    Connection,
    ConnectParams,
    MessageHandlers,
    InvalidStateError,
)
from ohmqtt.connection.fsm import FSM
from ohmqtt.connection.states import (
    ClosedState,
    ClosingState,
    ConnectedState,
    ConnectingState,
    ReconnectWaitState,
    ShutdownState,
)
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode
from ohmqtt.packet import (
    MQTTPublishPacket,
)


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=MessageHandlers)  # type: ignore[no-any-return]


@pytest.fixture
def mock_fsm(mocker: MockerFixture) -> Mock:
    mock_fsm = mocker.Mock(spec=FSM)
    mock_fsm.lock = mocker.MagicMock()
    mock_fsm.lock.__enter__.return_value = mock_fsm.lock
    mocker.patch("ohmqtt.connection.FSM", return_value=mock_fsm)
    return mock_fsm  # type: ignore[no-any-return]


def test_connection_handle_packet(mocker: MockerFixture) -> None:
    fake_handlers = [
        mocker.Mock(spec=Callable[[MQTTPublishPacket], None]),
        mocker.Mock(spec=Callable[[MQTTPublishPacket], None]),
    ]

    handlers = MessageHandlers()
    with handlers:
        for fake_handler in fake_handlers:
            handlers.register(MQTTPublishPacket, fake_handler)
    connection = Connection(handlers)
    packet = MQTTPublishPacket()

    connection.handle_packet(packet)
    for handler in fake_handlers:
        handler.assert_called_once_with(packet)
        handler.reset_mock()

    # At least one exception should be raised if any handler raises an exception
    fake_handlers[0].side_effect = Exception("TEST")
    with pytest.raises(Exception):
        connection.handle_packet(packet)
    # All handlers are still called
    for handler in fake_handlers:
        handler.assert_called_once_with(packet)
        handler.reset_mock()

    # MQTTError has priority over other exceptions
    fake_handlers[1].side_effect = MQTTError("TEST", MQTTReasonCode.UnspecifiedError)
    with pytest.raises(MQTTError):
        connection.handle_packet(packet)
    # All handlers are still called
    for handler in fake_handlers:
        handler.assert_called_once_with(packet)
        handler.reset_mock()


def test_connection_can_send(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.get_state.return_value = ClosedState
    assert not connection.can_send()
    mock_fsm.get_state.return_value = ConnectedState
    assert connection.can_send()


def test_connection_send(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.get_state.return_value = ConnectedState
    mock_fsm.env.packet_buffer = deque()
    packet = MQTTPublishPacket(topic="test/topic", payload=b"test")
    connection.send(packet)
    assert mock_fsm.env.packet_buffer.popleft() == packet
    assert not mock_fsm.env.packet_buffer  # Buffer should be empty now
    mock_fsm.selector.interrupt.assert_called_once_with()
    mock_fsm.get_state.return_value = ClosedState
    with pytest.raises(InvalidStateError):
        connection.send(packet)


def test_connection_connect(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    params = ConnectParams(address=Address("test_address"))
    connection.connect(params)
    mock_fsm.set_params.assert_called_once_with(params)
    mock_fsm.change_state.assert_called_once_with(ConnectingState)
    mock_fsm.selector.interrupt.assert_called_once_with()


def test_connection_disconnect(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    connection.disconnect()
    mock_fsm.request_state.assert_called_once_with(ClosingState)
    mock_fsm.selector.interrupt.assert_called_once_with()


def test_connection_shutdown(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    connection.shutdown()
    mock_fsm.request_state.assert_called_once_with(ShutdownState)
    mock_fsm.selector.interrupt.assert_called_once_with()


def test_connection_is_connected(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.get_state.return_value = ConnectedState
    assert connection.is_connected() is True
    mock_fsm.get_state.return_value = ClosedState
    assert connection.is_connected() is False


def test_connection_wait_for_connect(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.wait_for_state.return_value = True
    assert connection.wait_for_connect(0.1) is True
    mock_fsm.wait_for_state.assert_called_once_with((ConnectedState,), 0.1)
    mock_fsm.wait_for_state.return_value = False
    assert connection.wait_for_connect(0.1) is False


def test_connection_wait_for_disconnect(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.wait_for_state.return_value = True
    assert connection.wait_for_disconnect(0.1) is True
    mock_fsm.wait_for_state.assert_called_once_with((ClosedState, ShutdownState, ReconnectWaitState), 0.1)
    mock_fsm.wait_for_state.return_value = False
    assert connection.wait_for_disconnect(0.1) is False


def test_connection_wait_for_shutdown(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.wait_for_state.return_value = True
    assert connection.wait_for_shutdown(0.1) is True
    mock_fsm.wait_for_state.assert_called_once_with((ShutdownState,), 0.1)
    mock_fsm.wait_for_state.return_value = False
    assert connection.wait_for_shutdown(0.1) is False


def test_connection_loop_once(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    connection.loop_once(0.1)
    mock_fsm.loop_once.assert_called_once_with(0.1)
    connection.loop_once()
    # Default should be non-blocking.
    mock_fsm.loop_once.assert_called_with(0.0)


def test_connection_loop_forever(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    connection.loop_forever()
    mock_fsm.loop_until_state.assert_called_once_with((ShutdownState,))


def test_connection_loop_until_connected(mock_fsm: Mock, mock_handlers: Mock) -> None:
    connection = Connection(mock_handlers)
    mock_fsm.loop_until_state.return_value = True
    assert connection.loop_until_connected(0.1) is True
    mock_fsm.loop_until_state.assert_called_once_with((ConnectedState,), 0.1)
    mock_fsm.loop_until_state.return_value = False
    assert connection.loop_until_connected(0.1) is False



from .closed import ClosedState as ClosedState
from .closing import ClosingState as ClosingState
from .connected import ConnectedState as ConnectedState
from .connecting import ConnectingState as ConnectingState
from .mqtt_handshake_connack import MQTTHandshakeConnAckState as MQTTHandshakeConnAckState
from .mqtt_handshake_connect import MQTTHandshakeConnectState as MQTTHandshakeConnectState
from .reconnect_wait import ReconnectWaitState as ReconnectWaitState
from .shutdown import ShutdownState as ShutdownState
from .tls_handshake import TLSHandshakeState as TLSHandshakeState
from .websocket_handshake_request import WebsocketHandshakeRequestState as WebsocketHandshakeRequestState
from .websocket_handshake_response import WebsocketHandshakeResponseState as WebsocketHandshakeResponseState


# Hook up transitions.
ConnectingState.transitions_to = (ClosingState, ClosedState, ShutdownState, TLSHandshakeState, WebsocketHandshakeRequestState, MQTTHandshakeConnectState)
ConnectingState.request_from = (ClosedState, ReconnectWaitState)

TLSHandshakeState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnectState, WebsocketHandshakeRequestState)

WebsocketHandshakeRequestState.transitions_to = (ClosingState, ClosedState, ShutdownState, WebsocketHandshakeResponseState)

WebsocketHandshakeResponseState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnectState)

MQTTHandshakeConnectState.transitions_to = (ClosingState, ClosedState, ShutdownState, MQTTHandshakeConnAckState)

MQTTHandshakeConnAckState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectedState)

ConnectedState.transitions_to = (ClosingState, ClosedState, ShutdownState)

ClosingState.transitions_to = (ClosedState, ShutdownState)
ClosingState.request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    WebsocketHandshakeRequestState,
    WebsocketHandshakeResponseState,
    ConnectedState,
    ReconnectWaitState,
)

ClosedState.transitions_to = (ConnectingState, ShutdownState, ReconnectWaitState)

ReconnectWaitState.transitions_to = (ClosingState, ClosedState, ShutdownState, ConnectingState)

ShutdownState.request_from = (
    ConnectingState,
    TLSHandshakeState,
    MQTTHandshakeConnectState,
    MQTTHandshakeConnAckState,
    WebsocketHandshakeRequestState,
    WebsocketHandshakeResponseState,
    ConnectedState,
    ReconnectWaitState,
    ClosingState,
    ClosedState,
)



import argparse
import string

from .common import add_broker_arguments, get_client, Command, SubParsersT
from .. import Client, MQTTPublishPacket, RetainPolicy


class SubscribeCommand(Command):
    """Subscribe to a topic and print messages."""

    @classmethod
    def register(cls, sub_parsers: SubParsersT) -> None:
        parser = sub_parsers.add_parser(
            "subscribe", help="Subscribe to a topic and print messages"
        )
        add_broker_arguments(parser)
        parser.add_argument("topic", type=str, help="The topic to subscribe to")
        parser.add_argument(
            "--max_qos", type=int, choices=[0, 1, 2], default=2, help="The maximum QoS level for received messages"
        )
        parser.add_argument(
            "--no_retained", action="store_true", help="Do not receive retained messages"
        )
        parser.set_defaults(func=cls.execute)

    @classmethod
    def execute(cls, args: argparse.Namespace) -> None:
        topic = args.topic
        max_qos = args.max_qos
        retain_policy = RetainPolicy.NEVER if args.no_retained else RetainPolicy.ALWAYS

        def callback(_: Client, msg: MQTTPublishPacket) -> None:
            content = f"<{len(msg.payload)} bytes of binary data>"
            try:
                decoded = msg.payload.decode("utf-8")
                if all(c in string.printable for c in decoded):
                    content = decoded
            except UnicodeDecodeError:
                pass
            print(f"{msg.topic} : {content}")

        client = get_client(args)

        client.subscribe(
            topic,
            callback,
            max_qos=max_qos,
            retain_policy=retain_policy,
        )
        try:
            client.loop_forever()
        except KeyboardInterrupt:
            pass
        finally:
            client.shutdown()



from typing import Any, Callable, Sequence

import pytest

from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTPropertyId, MQTTQoS
from ohmqtt.packet import (
    decode_packet,
    decode_packet_from_parts,
    MQTTPacket,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    MQTTDisconnectPacket,
    MQTTAuthPacket,
)
from ohmqtt.property import (
    MQTTProperties,
    MQTTWillProps,
)


def extract_props(cls: type[MQTTPacket], data: Any, props_t: type[MQTTProperties] | None = None) -> MQTTProperties:
    """Helper for extracting properties from test data."""
    if props_t is None:
        props = cls.props_type()
    else:
        props = props_t()
    for prop in data:
        k = prop[0]
        prop_key = MQTTPropertyId[k]
        prop_value: Any
        if prop_key in (MQTTPropertyId.CorrelationData, MQTTPropertyId.AuthenticationData):
            prop_value = bytes.fromhex(prop[1])
        elif prop_key == MQTTPropertyId.SubscriptionIdentifier:
            prop_value = set(prop[1:])
        elif prop_key == MQTTPropertyId.UserProperty:
            prop_value = [tuple([pk, pv]) for pk, pv in prop[1:]]
        else:
            prop_value = prop[1]
        setattr(props, k, prop_value)
    return props


def extract_args(cls: type[MQTTPacket], data: dict[str, Any], binary_args: Sequence[str]) -> dict[str, Any]:
    """Helper for extracting arguments from test data."""
    args = {}
    for k, v in data.items():
        value: Any = v
        if k in binary_args:
            value = bytes.fromhex(v)
        elif k == "reason_code":
            value = MQTTReasonCode(v)
        elif k == "reason_codes":
            value = [MQTTReasonCode(x) for x in v]
        elif k == "properties":
            value = extract_props(cls, v)
        elif k == "will_props":
            value = extract_props(cls, v, MQTTWillProps)
        elif k == "qos":
            value = MQTTQoS(v)
        args[k] = value
    return args


def run_encode_cases(
    cls: type[MQTTPacket],
    test_data: list[dict[str, Any]],
    binary_args: Sequence[str] = tuple(),
    transform_args: Callable[[dict[str, Any]], None] | None = None
) -> None:
    for case in test_data:
        args = extract_args(cls, case["args"], binary_args)
        if transform_args is not None:
            transform_args(args)
        packet = cls(**args)
        encoded = packet.encode()
        assert encoded == bytes.fromhex(case["raw"]), encoded.hex()

        decoded = decode_packet(bytes.fromhex(case["raw"]))
        assert type(decoded) is cls
        for attr, value in args.items():
            assert getattr(decoded, attr) == value, f"{attr}: {getattr(decoded, attr)} != {value}"

        assert packet == packet
        assert packet == decoded
        assert packet != b"not a packet"  # type: ignore[comparison-overlap]
        assert str(packet) == str(decoded)
        assert repr(packet) == repr(decoded)
        assert not hasattr(packet, "__dict__")
        assert all(hasattr(packet, attr) for attr in packet.__slots__)
        with pytest.raises(TypeError):
            hash(packet)


def run_decode_alt_cases(
    cls: type[MQTTPacket],
    test_data: list[dict[str, Any]],
    binary_args: Sequence[str] = tuple(),
    transform_args: Callable[[dict[str, Any]], None] | None = None
) -> None:
    for case in test_data:
        args = extract_args(cls, case["args"], binary_args)
        if transform_args is not None:
            transform_args(args)

        decoded = decode_packet(bytes.fromhex(case["raw"]))
        assert type(decoded) is cls
        for attr, value in args.items():
            assert getattr(decoded, attr) == value, f"{attr}: {getattr(decoded, attr)} != {value}"


def run_decode_error_cases(test_data: list[dict[str, Any]]) -> None:
    for case in test_data:
        with pytest.raises(MQTTError) as excinfo:
            decode_packet(bytes.fromhex(case["raw"]))
        assert excinfo.value.reason_code == case["reason_code"]


def test_packet_connect_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTConnectPacket, test_data, ("password", "will_payload"))


def test_packet_connect_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_connack_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTConnAckPacket, test_data)


def test_packet_connack_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_publish_encode(test_data: list[dict[str, Any]]) -> None:
    def transform_args(args: dict[str, Any]) -> None:
        args["payload"] = args["payload"].encode("utf-8")
    run_encode_cases(MQTTPublishPacket, test_data, transform_args=transform_args)


def test_packet_publish_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_puback_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTPubAckPacket, test_data)


def test_packet_puback_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTPubAckPacket, test_data)


def test_packet_puback_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_pubrec_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTPubRecPacket, test_data)


def test_packet_pubrec_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTPubRecPacket, test_data)


def test_packet_pubrec_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_pubrel_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTPubRelPacket, test_data)


def test_packet_pubrel_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTPubRelPacket, test_data)


def test_packet_pubrel_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_pubcomp_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTPubCompPacket, test_data)


def test_packet_pubcomp_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTPubCompPacket, test_data)


def test_packet_pubcomp_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_subscribe_encode(test_data: list[dict[str, Any]]) -> None:
    def transform_args(args: dict[str, Any]) -> None:
        args["topics"] = [(t, int(q)) for t, q in args["topics"]]
    run_encode_cases(MQTTSubscribePacket, test_data, transform_args=transform_args)


def test_packet_subscribe_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_suback_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTSubAckPacket, test_data)


def test_packet_suback_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_unsubscribe_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTUnsubscribePacket, test_data)


def test_packet_unsubscribe_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_unsuback_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTUnsubAckPacket, test_data)


def test_packet_unsuback_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_pingreq_packet() -> None:
    raw = b"\xc0\x00"
    packet = MQTTPingReqPacket()
    encoded = packet.encode()
    assert encoded == raw

    decoded = decode_packet(raw)
    assert type(decoded) is MQTTPingReqPacket
    assert packet == decoded
    assert str(packet) == str(decoded)
    assert not hasattr(packet, "__dict__")
    with pytest.raises(TypeError):
        hash(packet)

    with pytest.raises(MQTTError):
        decode_packet(b"\xc0")
    with pytest.raises(MQTTError):
        decode_packet(b"\xc0\x01\x00")
    with pytest.raises(MQTTError):
        decode_packet(b"\xc1\x00")


def test_packet_pingresp_packet() -> None:
    raw = b"\xd0\x00"
    packet = MQTTPingRespPacket()
    encoded = packet.encode()
    assert encoded == raw

    decoded = decode_packet(raw)
    assert type(decoded) is MQTTPingRespPacket
    assert packet == decoded
    assert str(packet) == str(decoded)
    assert not hasattr(packet, "__dict__")
    with pytest.raises(TypeError):
        hash(packet)

    with pytest.raises(MQTTError):
        decode_packet(b"\xd0")
    with pytest.raises(MQTTError):
        decode_packet(b"\xd0\x01\x00")
    with pytest.raises(MQTTError):
        decode_packet(b"\xd1\x00")


def test_packet_disconnect_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTDisconnectPacket, test_data)


def test_packet_disconnect_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTDisconnectPacket, test_data)


def test_packet_disconnect_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_auth_encode(test_data: list[dict[str, Any]]) -> None:
    run_encode_cases(MQTTAuthPacket, test_data)


def test_packet_auth_decode_alt(test_data: list[dict[str, Any]]) -> None:
    run_decode_alt_cases(MQTTAuthPacket, test_data)


def test_packet_auth_decode_errors(test_data: list[dict[str, Any]]) -> None:
    run_decode_error_cases(test_data)


def test_packet_comparison() -> None:
    packet1 = MQTTConnectPacket("foo", 1)
    packet2 = MQTTConnectPacket("foo", 1)
    packet3 = MQTTConnectPacket("foo", 2)
    assert packet1 == packet2
    assert packet1 != packet3
    assert packet1 != "foo" # type: ignore[comparison-overlap]
    assert packet1 != 1  # type: ignore[comparison-overlap]
    assert packet1 is not None
    assert packet1 != object()


def test_packet_decode_packet_errors() -> None:
    for case in (b"\x00", b"\x20\x01"):
        with pytest.raises(MQTTError) as excinfo:
            decode_packet(case)
        assert excinfo.value.reason_code == MQTTReasonCode.MalformedPacket


def test_packet_decode_packet_from_parts_errors() -> None:
    with pytest.raises(MQTTError) as excinfo:
        decode_packet_from_parts(0, memoryview(b"\x00"))
    assert excinfo.value.reason_code == MQTTReasonCode.MalformedPacket



from __future__ import annotations

from typing import Final, TypeAlias

from .connection import Connection, ConnectParams, InvalidStateError, MessageHandlers
from .error import MQTTError
from .handles import PublishHandle
from .logger import get_logger
from .mqtt_spec import MAX_PACKET_ID, MQTTReasonCode, MQTTQoS
from .packet import (
    MQTTConnAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from .property import MQTTPublishProps
from .persistence.base import Persistence
from .persistence.in_memory import InMemoryPersistence
from .persistence.sqlite import SQLitePersistence
from .subscriptions import Subscriptions
from .topic_alias import AliasPolicy, OutboundTopicAlias

logger: Final = get_logger("session")

SessionSendablePacketT: TypeAlias = (
    MQTTPublishPacket |
    MQTTPubAckPacket |
    MQTTPubRecPacket |
    MQTTPubRelPacket |
    MQTTPubCompPacket
)


class Session:
    __slots__ = (
        "_cond",
        "connection",
        "inflight",
        "params",
        "persistence",
        "server_receive_maximum",
        "subscriptions",
        "topic_alias",
    )
    persistence: Persistence

    def __init__(
        self,
        handlers: MessageHandlers,
        subscriptions: Subscriptions,
        connection: Connection,
        *,
        db_path: str = "",
        db_fast: bool = False,
    ) -> None:
        self._cond = connection.fsm.cond
        self.topic_alias = OutboundTopicAlias()
        self.inflight = 0
        self.params = ConnectParams()
        self.subscriptions = subscriptions
        self.connection = connection
        if db_path:
            self.persistence = SQLitePersistence(db_path, db_fast=db_fast)
        else:
            self.persistence = InMemoryPersistence()
        self.server_receive_maximum = 0

        handlers.register(MQTTConnAckPacket, self.handle_connack)
        handlers.register(MQTTPublishPacket, self.handle_publish)
        handlers.register(MQTTPubAckPacket, self.handle_puback)
        handlers.register(MQTTPubRecPacket, self.handle_pubrec)
        handlers.register(MQTTPubRelPacket, self.handle_pubrel)
        handlers.register(MQTTPubCompPacket, self.handle_pubcomp)

    def set_params(self, params: ConnectParams) -> None:
        with self._cond:
            self.params = params

    def publish(
        self,
        topic: str,
        payload: bytes,
        *,
        qos: MQTTQoS = MQTTQoS.Q0,
        retain: bool = False,
        properties: MQTTPublishProps | None = None,
        alias_policy: AliasPolicy = AliasPolicy.NEVER,
    ) -> PublishHandle:
        """Publish a message to a topic."""
        properties = properties if properties is not None else MQTTPublishProps()
        if qos > MQTTQoS.Q0:
            with self._cond:
                handle = self.persistence.add(
                    topic=topic,
                    payload=payload,
                    qos=qos,
                    retain=retain,
                    properties=properties,
                    alias_policy=alias_policy,
                )
                if self.connection.can_send():
                    self._flush()
                else:
                    logger.debug("Connection is not ready to send, deferring flush")
                return handle
        else:
            packet = MQTTPublishPacket(
                topic=topic,
                payload=payload,
                qos=qos,
                retain=retain,
                properties=properties,
            )
            try:
                self._send_aliased(packet, alias_policy)
            except InvalidStateError:
                # Not being able to send a packet is not an error for QoS 0.
                logger.debug("Failed to send QoS 0 packet (invalid connection state), ignoring")
            handle = PublishHandle(self._cond)
            handle.exc = ValueError("QoS 0 messages will not be acknowledged")
            return handle

    def _send_packet(self, packet: SessionSendablePacketT) -> None:
        """Try to send a packet to the server."""
        self.connection.send(packet)

    def _send_aliased(
        self,
        packet: MQTTPublishPacket,
        alias_policy: AliasPolicy,
    ) -> None:
        """Send a potentially topic aliased PUBLISH packet to the server.

        This will handle topic aliasing if the server supports it."""
        if alias_policy == AliasPolicy.NEVER:
            # Fast path for no aliasing.
            self._send_packet(packet)
            return
        with self._cond:
            topic = packet.topic
            lookup = self.topic_alias.lookup(topic, alias_policy)
            assert lookup.alias > 0
            packet.properties.TopicAlias = lookup.alias
            if lookup.existed:
                packet.topic = ""
            try:
                self._send_packet(packet)
            except InvalidStateError:
                if not lookup.existed:
                    self.topic_alias.pop()
                raise

    def handle_connack(self, packet: MQTTConnAckPacket) -> None:
        """Handle a connection open event."""
        if packet.reason_code.is_error():
            logger.error("CONNACK error code: %s", packet.reason_code)
            raise MQTTError("Connection failed", packet.reason_code)
        with self._cond:
            self.inflight = 0
            self.topic_alias.reset()
            if packet.properties.ReceiveMaximum is not None:
                self.server_receive_maximum = packet.properties.ReceiveMaximum
            else:
                self.server_receive_maximum = MAX_PACKET_ID - 1
            if packet.properties.TopicAliasMaximum is not None:
                self.topic_alias.max_alias = packet.properties.TopicAliasMaximum
            else:
                self.topic_alias.max_alias = 0
            if packet.properties.AssignedClientIdentifier:
                client_id = packet.properties.AssignedClientIdentifier
            else:
                client_id = self.params.client_id
            if not client_id:
                raise MQTTError("No client ID provided", MQTTReasonCode.ProtocolError)
            clear_persistence = not packet.session_present
            self.persistence.open(client_id, clear=clear_persistence)
            self.subscriptions.handle_connack(packet)
            self._flush()

    def handle_publish(self, packet: MQTTPublishPacket) -> None:
        """Handle a PUBLISH packet from the server."""
        if packet.qos == MQTTQoS.Q2 and not self.persistence.check_rec(packet):
            logger.debug("Received duplicate QoS 2 packet with ID %d", packet.packet_id)
        else:
            self.subscriptions.handle_publish(packet)

        if packet.qos == MQTTQoS.Q1:
            ack_packet = MQTTPubAckPacket(packet_id=packet.packet_id)
            self._send_packet(ack_packet)
        elif packet.qos == MQTTQoS.Q2:
            self.persistence.set_rec(packet)
            rec_packet = MQTTPubRecPacket(packet_id=packet.packet_id)
            self._send_packet(rec_packet)

    def handle_puback(self, packet: MQTTPubAckPacket) -> None:
        """Handle a PUBACK packet from the server."""
        if packet.reason_code.is_error():
            logger.error("Received PUBACK with error code: %s", packet.reason_code)
        with self._cond:
            self.persistence.ack(packet)
            self.inflight -= 1
            self._flush()

    def handle_pubrec(self, packet: MQTTPubRecPacket) -> None:
        """Handle a PUBREC packet from the server."""
        with self._cond:
            if packet.reason_code.is_error():
                logger.error("Received PUBREC with error code: %s", packet.reason_code)
            self.persistence.ack(packet)
            self.inflight -= 1
            self._flush()

    def handle_pubrel(self, packet: MQTTPubRelPacket) -> None:
        """Handle a PUBREL packet from the server."""
        if packet.reason_code.is_error():
            logger.error("Received PUBREL with error code: %s", packet.reason_code)
        self.persistence.rel(packet)
        comp_packet = MQTTPubCompPacket(packet_id=packet.packet_id)
        self._send_packet(comp_packet)

    def handle_pubcomp(self, packet: MQTTPubCompPacket) -> None:
        """Handle a PUBCOMP packet from the server."""
        if packet.reason_code.is_error():
            logger.error("Received PUBCOMP with error code: %s", packet.reason_code)
        with self._cond:
            self.persistence.ack(packet)
            self.inflight -= 1
            self._flush()

    def _flush(self) -> None:
        """Send queued packets up to the server's receive maximum."""
        with self._cond:
            allowed_count = self.server_receive_maximum - self.inflight
            pending_message_ids = self.persistence.get(allowed_count)
            for message_id in pending_message_ids:
                packet, alias_policy = self.persistence.render(message_id)
                self.inflight += 1
                try:
                    if type(packet) is MQTTPublishPacket:
                        self._send_aliased(packet, alias_policy)
                    else:
                        self._send_packet(packet)
                except InvalidStateError:
                    # A race during transition out of ConnectedState can cause this.
                    # The rest of the machinery should be able to handle it.
                    logger.debug("Failed to send reliable packet (invalid connection state), ignoring")



import threading
from typing import ClassVar, Generic, TypeAlias, TypeVar
import weakref

from .packet import (
    MQTTPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubCompPacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
)

_T = TypeVar("_T", bound=MQTTPacket)

class Handle(Generic[_T]):
    __slots__: ClassVar[tuple[str, ...]] = ("__weakref__", "_cond", "ack", "exc")
    ack: _T | None
    exc: Exception | None
    _cond: weakref.ref[threading.Condition]

    def __init__(self, cond: threading.Condition) -> None:
        self.ack = None
        self.exc = None
        self._cond = weakref.ref(cond)

    def wait_for_ack(self, timeout: float | None = None) -> _T:
        """Wait for the ack packet or raise an exception if the operation failed.

        :raise TimeoutError: Acknowledgement timed out.
        :return: The final acknowledgement packet."""
        cond = self._cond()
        if cond is None:
            raise RuntimeError("Condition variable is no longer available")
        with cond:
            if self.ack is None and self.exc is None:
                cond.wait_for(lambda: self.ack is not None or self.exc is not None, timeout=timeout)
            if self.exc is not None:
                raise self.exc
            if self.ack is None:
                raise TimeoutError("Timeout waiting for ack")
            return self.ack


PublishHandle: TypeAlias = Handle[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket]
SubscribeHandle: TypeAlias = Handle[MQTTSubAckPacket]
UnsubscribeHandle: TypeAlias = Handle[MQTTUnsubAckPacket]



import threading

import pytest

from ohmqtt.handles import (
    PublishHandle,
    SubscribeHandle,
    UnsubscribeHandle,
)
from ohmqtt.packet import (
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubCompPacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
)


@pytest.mark.parametrize("handle_cls", [PublishHandle, SubscribeHandle, UnsubscribeHandle])
def test_handles_slots(handle_cls: type[PublishHandle | SubscribeHandle | UnsubscribeHandle]) -> None:
    """Test the UnreliablePublishHandle class."""
    cond = threading.Condition()
    handle = handle_cls(cond)

    assert not hasattr(handle, "__dict__")
    assert all(hasattr(handle, attr) for attr in handle.__slots__)


def _test_ack(
    handle_cls: type[PublishHandle | SubscribeHandle | UnsubscribeHandle],
    ack_cls: type[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket | MQTTSubAckPacket | MQTTUnsubAckPacket],
) -> None:
    start = threading.Event()
    cond = threading.Condition()
    handle = PublishHandle(cond)
    ack = ack_cls(packet_id=1)
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(timeout=0.001)

    def do_ack() -> None:
        start.wait()
        with cond:
            handle.ack = ack  # type: ignore[assignment]
            cond.notify_all()
    thread = threading.Thread(target=do_ack)
    thread.start()

    with cond:
        start.set()
        assert handle.wait_for_ack(timeout=0.01) == ack

    thread.join()


def _test_exc(
    handle_cls: type[PublishHandle | SubscribeHandle | UnsubscribeHandle],
    ack_cls: type[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket | MQTTSubAckPacket | MQTTUnsubAckPacket],
) -> None:
    start = threading.Event()
    cond = threading.Condition()
    handle = PublishHandle(cond)
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(timeout=0.001)

    def do_ack() -> None:
        start.wait()
        with cond:
            handle.exc = ValueError("TEST")
            cond.notify_all()
    thread = threading.Thread(target=do_ack)
    thread.start()

    with cond:
        start.set()
        with pytest.raises(ValueError, match="TEST"):
            handle.wait_for_ack(timeout=0.01)

    thread.join()


@pytest.mark.parametrize("ack_cls", [MQTTPubAckPacket, MQTTPubRecPacket, MQTTPubCompPacket])
def test_handles_publish_ack(ack_cls: type[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket]) -> None:
    _test_ack(PublishHandle, ack_cls)


@pytest.mark.parametrize("ack_cls", [MQTTPubAckPacket, MQTTPubRecPacket, MQTTPubCompPacket])
def test_handles_publish_exc(ack_cls: type[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket]) -> None:
    _test_exc(PublishHandle, ack_cls)


def test_handles_subscribe_ack() -> None:
    _test_ack(SubscribeHandle, MQTTSubAckPacket)


def test_handles_subscribe_exc() -> None:
    _test_exc(SubscribeHandle, MQTTSubAckPacket)


def test_handles_unsubscribe_ack() -> None:
    _test_ack(UnsubscribeHandle, MQTTUnsubAckPacket)


def test_handles_unsubscribe_exc() -> None:
    _test_exc(UnsubscribeHandle, MQTTUnsubAckPacket)


@pytest.mark.parametrize("handle_cls", [PublishHandle, SubscribeHandle, UnsubscribeHandle])
def test_handles_deref(handle_cls: type[PublishHandle | SubscribeHandle | UnsubscribeHandle]) -> None:
    cond = threading.Condition()
    handle = handle_cls(cond)
    del cond
    with pytest.raises(RuntimeError, match="Condition variable is no longer available"):
        handle.wait_for_ack(timeout=0.001)



import socket
import time

import pytest

from .util.fake_broker import FakeBroker, FakeWebsocketBroker
from ohmqtt.connection.decoder import IncrementalDecoder
from ohmqtt.connection.states.websocket_handshake_response import WebsocketHandshakeResponseState
from ohmqtt.connection.wslib import (
    OpCode,
    frame_ws_data,
    deframe_ws_data,
    generate_nonce,
)
from ohmqtt.mqtt_spec import MQTTReasonCode, MQTTQoS
from ohmqtt.packet import (
    decode_packet,
    MQTTPacket,
    MQTTConnectPacket,
    MQTTConnAckPacket,
    MQTTSubscribePacket,
    MQTTSubAckPacket,
    MQTTUnsubscribePacket,
    MQTTUnsubAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTPingReqPacket,
    MQTTPingRespPacket,
    MQTTDisconnectPacket,
)


def send_to_broker(sock: socket.socket, broker: FakeBroker, packet: MQTTPacket) -> None:
    sock.sendall(packet.encode())
    time.sleep(0.05)
    assert len(broker.received) == 1
    assert broker.received.popleft() == packet


def send_to_broker_ws(sock: socket.socket, broker: FakeBroker, packet: MQTTPacket) -> None:
    sock.sendall(frame_ws_data(OpCode.BINARY, packet.encode()))
    time.sleep(0.05)
    assert len(broker.received) == 1
    assert broker.received.popleft() == packet


def expect_from_broker(sock: socket.socket, decoder: IncrementalDecoder, packet: MQTTPacket) -> None:
    pkt = decoder.decode(sock)
    assert pkt == packet


def expect_from_broker_ws(sock: socket.socket, buffer: bytearray, packet: MQTTPacket) -> None:
    buffer.extend(sock.recv(0xffff))
    frame = deframe_ws_data(buffer)
    assert frame is not None
    opcode, payload, masked, frame_length = frame
    assert not masked
    assert opcode == OpCode.BINARY
    del buffer[:frame_length]
    pkt = decode_packet(payload)
    assert pkt == packet


def test_fake_broker() -> None:
    decoder = IncrementalDecoder()
    with FakeBroker() as broker:
        assert broker.port > 0
        assert len(broker.received) == 0

        sock = socket.create_connection(("localhost", broker.port))
        sock.settimeout(0.05)

        # CONNECT
        connect_packet = MQTTConnectPacket(
            client_id="test_client",
            clean_start=True,
        )
        send_to_broker(sock, broker, connect_packet)
        assert broker.sock is not None
        expect_from_broker(sock, decoder, MQTTConnAckPacket())

        # SUBSCRIBE
        subscribe_packet = MQTTSubscribePacket(
            topics=[("test/topic", 2)],
            packet_id=1,
        )
        send_to_broker(sock, broker, subscribe_packet)
        expect_from_broker(sock, decoder, MQTTSubAckPacket(packet_id=1, reason_codes=[MQTTReasonCode.Success]))
        # Subscribe to another topic with no_local
        subscribe_packet = MQTTSubscribePacket(
            topics=[("test/topic2", 2 | 0x04)],
            packet_id=2,
        )
        send_to_broker(sock, broker, subscribe_packet)
        expect_from_broker(sock, decoder, MQTTSubAckPacket(packet_id=2, reason_codes=[MQTTReasonCode.Success]))

        # PUBLISH
        publish_packet_qos0 = MQTTPublishPacket(
            topic="test/topic",
            payload=b"Hello, MQTT!",
        )
        send_to_broker(sock, broker, publish_packet_qos0)
        expect_from_broker(sock, decoder, publish_packet_qos0)
        # Make sure we don't get the publish back on the no_local topic
        send_to_broker(sock, broker, MQTTPublishPacket(topic="test/topic2", payload=b"Hello, no_local!"))
        with pytest.raises(TimeoutError):
            sock.recv(1024)

        publish_packet_qos1 = MQTTPublishPacket(
            topic="test/topic",
            payload=b"Hello, MQTT QoS 1!",
            qos=MQTTQoS.Q1,
            packet_id=1,
        )
        send_to_broker(sock, broker, publish_packet_qos1)
        expect_from_broker(sock, decoder, publish_packet_qos1)
        expect_from_broker(sock, decoder, MQTTPubAckPacket(packet_id=1))

        publish_packet_qos2 = MQTTPublishPacket(
            topic="test/topic",
            payload=b"Hello, MQTT QoS 2!",
            qos=MQTTQoS.Q2,
            packet_id=2,
        )
        send_to_broker(sock, broker, publish_packet_qos2)
        expect_from_broker(sock, decoder, publish_packet_qos2)
        expect_from_broker(sock, decoder, MQTTPubRecPacket(packet_id=2))
        send_to_broker(sock, broker, MQTTPubRelPacket(packet_id=2))
        expect_from_broker(sock, decoder, MQTTPubCompPacket(packet_id=2))

        # UNSUBSCRIBE
        unsubscribe_packet = MQTTUnsubscribePacket(
            topics=["test/topic"],
            packet_id=3,
        )
        send_to_broker(sock, broker, unsubscribe_packet)
        expect_from_broker(sock, decoder, MQTTUnsubAckPacket(packet_id=3, reason_codes=[MQTTReasonCode.Success]))
        # Make sure we don't get the publish back
        send_to_broker(sock, broker, publish_packet_qos0)
        with pytest.raises(TimeoutError):
             sock.recv(1024)

        # PINGREQ
        pingreq_packet = MQTTPingReqPacket()
        send_to_broker(sock, broker, pingreq_packet)
        expect_from_broker(sock, decoder, MQTTPingRespPacket())

        # Send a message originating from the broker.
        broker.send(MQTTPingReqPacket())
        expect_from_broker(sock, decoder, MQTTPingReqPacket())

        # DISCONNECT
        disconnect_packet = MQTTDisconnectPacket(reason_code=MQTTReasonCode.NormalDisconnection)
        send_to_broker(sock, broker, disconnect_packet)
        assert sock.recv(1024) == b""


def test_fake_broker_websocket() -> None:
    with FakeWebsocketBroker() as broker:
        assert broker.port > 0
        assert len(broker.received) == 0

        sock = socket.create_connection(("localhost", broker.port))
        sock.settimeout(0.05)
        nonce = generate_nonce()
        buffer = bytearray()

        # Handshake
        handshake_request = (
            "GET / HTTP/1.1\r\n"
            "Host: localhost\r\n"
            "Upgrade: websocket\r\n"
            "Connection: Upgrade\r\n"
            f"Sec-WebSocket-Key: {nonce}\r\n"
            "Sec-WebSocket-Version: 13\r\n"
            "Sec-WebSocket-Protocol: mqtt\r\n"
            "\r\n"
        )
        sock.sendall(handshake_request.encode("utf-8"))
        handshake_response = sock.recv(0xffff)
        assert WebsocketHandshakeResponseState.validate_response(nonce, bytearray(handshake_response)) is True

        # CONNECT
        connect_packet = MQTTConnectPacket(
            client_id="test_client",
            clean_start=True,
        )
        send_to_broker_ws(sock, broker, connect_packet)
        assert broker.sock is not None
        expect_from_broker_ws(sock, buffer, MQTTConnAckPacket())

        # PINGREQ
        pingreq_packet = MQTTPingReqPacket()
        send_to_broker_ws(sock, broker, pingreq_packet)
        expect_from_broker_ws(sock, buffer, MQTTPingRespPacket())

        # Send a message originating from the broker.
        broker.send(MQTTPingReqPacket())
        expect_from_broker_ws(sock, buffer, MQTTPingReqPacket())

        # DISCONNECT
        disconnect_packet = MQTTDisconnectPacket(reason_code=MQTTReasonCode.NormalDisconnection)
        send_to_broker_ws(sock, broker, disconnect_packet)
        assert sock.recv(1024) == b""



from __future__ import annotations

from dataclasses import dataclass, field
from enum import IntEnum
from typing import NamedTuple

from .error import MQTTError
from .mqtt_spec import MQTTReasonCode
from .packet import MQTTPublishPacket


class MaxOutboundAliasError(Exception):
    """Exception raised when the maximum number of topic aliases is reached."""


class AliasPolicy(IntEnum):
    """Topic alias policy.

    NEVER: Never use topic aliases.

    TRY: Use topic aliases if possible.
        If an alias does not exist, attempt to create a new one.
        If the maximum number of aliases is reached, an alias will not be used.

    ALWAYS: Always use topic aliases.
        If an alias does not exist, attempt to create a new one.
        If the maximum number of aliases is reached, an exception will be raised."""
    NEVER = 0
    TRY = 1
    ALWAYS = 2


class OutboundLookupResult(NamedTuple):
    """Result of an outbound lookup.

    If the alias is 0, a topic alias should not be used in the publish packet.

    If the alias is not 0 and existed is False,
        the alias and topic should both be sent in the publish packet.

    If the alias is not 0 and existed is True,
        the alias should be sent in the publish packet and the topic should not be sent."""
    alias: int
    existed: bool


@dataclass(slots=True)
class InboundTopicAlias:
    """Inbound topic alias store."""
    aliases: dict[int, str] = field(default_factory=dict)
    max_alias: int = field(default=0, init=False)

    def reset(self) -> None:
        """Reset the topic alias state.

        This should be called at least when the connection is opened."""
        self.aliases.clear()
        self.max_alias = 0

    def handle(self, packet: MQTTPublishPacket) -> None:
        """Handle the topic alias in a publish packet.

        Stores the topic alias in the store if it is not recognized.

        If the topic alias is recognized, the topic is replaced with the stored topic."""
        topic = packet.topic
        alias = packet.properties.TopicAlias
        if alias is None:
            if not topic:
                raise MQTTError(
                    "Topic alias not found and topic is empty",
                    reason_code=MQTTReasonCode.ProtocolError,
                )
            return
        if alias > self.max_alias:
            raise MQTTError(
                f"Topic alias {alias} out of range",
                reason_code=MQTTReasonCode.TopicAliasInvalid,
            )

        if alias in self.aliases:
            if topic and self.aliases[alias] != topic:
                raise MQTTError(
                    f"Topic alias {alias} already exists",
                    reason_code=MQTTReasonCode.TopicAliasInvalid,
                )
            packet.topic = self.aliases[alias]
        else:
            if not topic:
                raise MQTTError(
                    "Topic alias not found and topic is empty",
                    reason_code=MQTTReasonCode.TopicAliasInvalid,
                )
            self.aliases[alias] = topic


@dataclass(slots=True)
class OutboundTopicAlias:
    """Inbound topic alias store."""
    aliases: dict[str, int] = field(default_factory=dict)
    next_alias: int = field(default=1, init=False)
    max_alias: int = field(default=0, init=False)

    def reset(self) -> None:
        """Reset the topic alias state.

        This should be called at least when the connection is opened."""
        self.aliases.clear()
        # Do not reset max_alias, as it is set upon connection.
        self.next_alias = 1

    def lookup(self, topic: str, policy: AliasPolicy) -> OutboundLookupResult:
        """Get the topic alias for a given topic from the client.

        An alias integer and a boolean indicating if the alias already existed will be returned.

        If the alias integer is 0, the alias was not created and the topic is not in the store.
        In this case, the topic alias must not be used in the publish packet."""
        if policy == AliasPolicy.NEVER:
            return OutboundLookupResult(0, False)
        if topic in self.aliases:
            return OutboundLookupResult(self.aliases[topic], True)
        if self.next_alias > self.max_alias:
            if policy == AliasPolicy.ALWAYS:
                raise MaxOutboundAliasError("Out of topic aliases and policy is ALWAYS")
            return OutboundLookupResult(0, False)
        alias = self.next_alias
        self.aliases[topic] = alias
        self.next_alias += 1
        return OutboundLookupResult(alias, False)

    def pop(self) -> None:
        """Remove the last assigned alias."""
        self.aliases.pop(next(reversed(self.aliases)), None)
        self.next_alias -= 1



import socket
import sys
from typing import cast, Final


HAS_AF_UNIX: Final = sys.platform != "win32"

if sys.platform == "win32":
    AF_UNIX: Final = cast(socket.AddressFamily, 1)
else:
    AF_UNIX: Final = socket.AF_UNIX


class PlatformError(Exception):
    """A feature is not supported on this platform."""



from __future__ import annotations

from functools import wraps
import sys
import threading
from typing import Any, Callable, Concatenate, ParamSpec, TypeAlias, TypeVar, Union

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


LockLike: TypeAlias = Union[threading.RLock, threading.Lock]
ProtectedT = TypeVar("ProtectedT", bound="Protected")
ProtectP = ParamSpec("ProtectP")
ProtectR = TypeVar("ProtectR")


def protect(
    func: Callable[Concatenate[ProtectedT, ProtectP], ProtectR],
) -> Callable[Concatenate[ProtectedT, ProtectP], ProtectR]:
    """Decorator to protect a method of a Protected instance."""
    @wraps(func)
    def wrapper(self: ProtectedT, /, *args: Any, **kwargs: Any) -> ProtectR:
        if not self._is_owned():
            raise RuntimeError(f"{self.__class__.__name__} instance lock is not owned by this thread")
        return func(self, *args, **kwargs)
    return wrapper


class Protected:
    """A wrapper to protect a resource or resources.

    Combine with the `@protect` decorator to protect methods of this class."""
    __slots__ = ("__weakref__", "_is_owned", "_lock", "acquire", "release")

    def __init__(self, lock: LockLike | None = None) -> None:
        self._lock = threading.RLock() if lock is None else lock
        self.acquire = self._lock.acquire
        self.release = self._lock.release

        if hasattr(self._lock, "_is_owned"):
            self._is_owned = self._lock._is_owned  # noqa: SLF001
        else:
            raise RuntimeError(f"{self.__class__.__name__} lock does not support _is_owned() method")

    def __enter__(self) -> Self:
        self.acquire()
        return self

    def __exit__(self, *args: object) -> None:
        self.release()



from __future__ import annotations

import ssl
from typing import cast, get_args, Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .closing import ClosingState
from ..decoder import ClosedSocketError
from ..types import ConnectParams, ReceivablePacketT, StateData, StateEnvironment
from ..wslib import OpCode, frame_ws_data, WebsocketError
from ...error import MQTTError
from ...logger import get_logger
from ...mqtt_spec import MQTTPacketType, MQTTReasonCode
from ...packet import decode_packet, MQTTPacket, MQTTPingReqPacket, MQTTPingRespPacket

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.connected")


class ConnectedState(FSMState):
    """Connected to the broker. Full duplex messaging in this state."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.keepalive.mark_init()
        env.packet_buffer.clear()
        state_data.write_buffer.clear()
        assert state_data.connack is not None, "Got to ConnectedState without a CONNACK"
        env.packet_callback(state_data.connack)
        state_data.open_called = True

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.keepalive.should_close():
            logger.error("Keepalive timeout, closing socket")
            fsm.change_state(ClosedState)
            return True

        if state_data.keepalive.should_send_ping():
            logger.debug("---> PING")
            env.packet_buffer.append(MQTTPingReqPacket())
            state_data.keepalive.mark_ping()

        with fsm.selector:
            timeout = state_data.keepalive.get_next_timeout(max_wait)
            write_check = bool(state_data.write_buffer or env.packet_buffer)
            readable, writable = fsm.selector.select(read=True, write=write_check, timeout=timeout)

        # Batch encode outgoing packets until we have at least 64kB of data or the queue is drained.
        while env.packet_buffer and len(state_data.write_buffer) < 0xffff:
            packet = env.packet_buffer.popleft()
            if params.address.is_websocket():
                ws_frame = frame_ws_data(OpCode.BINARY, packet.encode())
                state_data.write_buffer.extend(ws_frame)
            else:
                state_data.write_buffer.extend(packet.encode())

        if writable:
            try:
                sent = state_data.sock.send(state_data.write_buffer)
                del state_data.write_buffer[:sent]
                state_data.keepalive.mark_send()
            except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
                pass
            except (BrokenPipeError, ConnectionResetError) as exc:
                logger.error("MQTT connection was closed: %s", exc)
                fsm.change_state(ClosedState)
                return True

        # Read one incoming packet at a time.
        if readable:
            try:
                cls.read_packet(fsm, state_data, env, params)
            except ClosedSocketError:
                logger.debug("Connection closed")
                fsm.change_state(ClosedState)
                return True
            except WebsocketError as exc:
                logger.error("WebSocket protocol error while connected: %s", exc)
                fsm.change_state(ClosedState)
                return True
            except MQTTError as exc:
                logger.error("There was a problem with data from broker, closing connection: %s", exc)
                state_data.disconnect_rc = exc.reason_code
                fsm.change_state(ClosedState)
                return True

        return False

    @classmethod
    def pull_tcp(cls, state_data: StateData) -> MQTTPacket | None:
        """Read an incoming packet from the TCP socket.

        Returns an MQTTPacket if one was read, or None if no complete MQTT packet was available."""
        return state_data.decoder.decode(state_data.sock)

    @classmethod
    def pull_ws(cls, state_data: StateData) -> MQTTPacket | None:
        """Read an incoming packet from the WebSocket.

        Non-data WebSocket frames (PING, PONG, CLOSE) are handled here.

        Returns an MQTTPacket if one was read, or None if no complete MQTT packet was available."""
        decode_result = state_data.ws_decoder.decode(state_data.sock)
        if decode_result is None:
            # No complete WebSocket frame available yet.
            return None
        opcode, payload = decode_result
        if opcode == OpCode.PING:
            logger.debug("<--- WEBSOCKET PING PONG --->")
            pong_frame = frame_ws_data(OpCode.PONG, payload)
            state_data.write_buffer.extend(pong_frame)
            return None
        if opcode == OpCode.CLOSE:
            raise ClosedSocketError("WebSocket connection closed by peer")
        if opcode != OpCode.BINARY:
            raise WebsocketError(f"Unexpected WebSocket opcode {opcode.name} while connected")
        return decode_packet(payload)

    @classmethod
    def read_packet(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        """Called by the underlying SocketWrapper when the socket is ready to read.

        Incrementally reads and decodes a packet from the socket.
        Complete packets are passed up to the read callback."""
        if params.address.is_websocket():
            packet = cls.pull_ws(state_data)
        else:
            packet = cls.pull_tcp(state_data)
        if packet is None:
            # Incomplete or handled internally.
            return

        # Ping requests and responses are handled at this layer.
        if packet.packet_type == MQTTPacketType.PINGRESP:
            logger.debug("<--- PONG")
            state_data.keepalive.mark_pong()
        elif packet.packet_type == MQTTPacketType.PINGREQ:
            logger.debug("<--- PING PONG --->")
            env.packet_buffer.append(MQTTPingRespPacket())
        elif packet.packet_type == MQTTPacketType.DISCONNECT:
            logger.debug("<--- %s", packet)
            logger.info("Broker sent DISCONNECT, closing connection")
            fsm.change_state(ClosingState)
        elif not isinstance(packet, get_args(ReceivablePacketT)):
            # To cast later, we must handle the exceptional cases at runtime.
            raise MQTTError("Unexpected packet type", reason_code=MQTTReasonCode.ProtocolError)
        else:
            # All other packets are passed to the read callback.
            env.packet_callback(cast(ReceivablePacketT, packet))



from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection import Connection, ConnectParams, InvalidStateError, MessageHandlers
from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTQoS, MQTTReasonCode
from ohmqtt.packet import (
    MQTTConnAckPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ohmqtt.property import MQTTConnAckProps, MQTTPublishProps
from ohmqtt.session import Session
from ohmqtt.subscriptions import Subscriptions
from ohmqtt.topic_alias import AliasPolicy


@pytest.fixture
def mock_handlers(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=MessageHandlers)  # type: ignore[no-any-return]


@pytest.fixture
def mock_connection(mocker: MockerFixture) -> Mock:
    conn = mocker.MagicMock(spec=Connection)
    conn.fsm.cond.__enter__.return_value = mocker.Mock()
    conn.fsm.lock.__enter__.return_value = mocker.Mock()
    return conn  # type: ignore[no-any-return]


@pytest.fixture
def mock_subscriptions(mocker: MockerFixture) -> Mock:
    return mocker.Mock(spec=Subscriptions)  # type: ignore[no-any-return]


def test_session_publish_qos0(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload")
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    ))
    mock_connection.send.reset_mock()
    assert handle.ack is None
    with pytest.raises(ValueError, match="QoS 0 messages will not be acknowledged"):
        handle.wait_for_ack(timeout=1)


def test_session_publish_qos0_aliased(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(TopicAliasMaximum=255)))
    mock_connection.can_send.return_value = False
    mock_connection.send.side_effect = InvalidStateError("TEST")

    session.publish("test/topic", b"test payload", alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    mock_connection.can_send.return_value = True
    mock_connection.send.side_effect = None

    session.publish("test/topic", b"test payload", alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    for _ in range(3):
        session.publish("test/topic", b"test payload", alias_policy=AliasPolicy.TRY)
        mock_connection.send.assert_called_with(MQTTPublishPacket(
            topic="",
            payload=b"test payload",
            properties=MQTTPublishProps(TopicAlias=1),
        ))
        mock_connection.send.reset_mock()


def test_session_publish_qos0_unconnected(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = False
    mock_connection.send.side_effect = InvalidStateError("TEST")

    handle = session.publish("test/topic", b"test payload")
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    ))
    mock_connection.send.reset_mock()
    assert handle.ack is None
    with pytest.raises(ValueError, match="QoS 0 messages will not be acknowledged"):
        handle.wait_for_ack(0.001)


@pytest.mark.parametrize("ack_rc", [MQTTReasonCode.Success, MQTTReasonCode.UnspecifiedError])
def test_session_publish_qos1(ack_rc: MQTTReasonCode, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q1)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q1,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    ack = MQTTPubAckPacket(packet_id=1, reason_code=ack_rc)
    session.handle_puback(ack)
    if ack_rc.is_error():
        with pytest.raises(MQTTError) as excinfo:
            handle.wait_for_ack(0.001)
        assert excinfo.value.reason_code == ack_rc
    else:
        assert handle.wait_for_ack() is ack


def test_session_publish_qos1_unconnected(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = False
    mock_connection.send.side_effect = InvalidStateError("TEST")

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q1)
    mock_connection.send.assert_not_called()
    assert handle.ack is None
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)


@pytest.mark.parametrize("comp_rc", [MQTTReasonCode.Success, MQTTReasonCode.MessageRateTooHigh])
def test_session_publish_qos2(comp_rc: MQTTReasonCode, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q2)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q2,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    session.handle_pubrec(MQTTPubRecPacket(packet_id=1))
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    mock_connection.send.assert_called_with(MQTTPubRelPacket(packet_id=1))
    mock_connection.send.reset_mock()

    ack = MQTTPubCompPacket(packet_id=1, reason_code=comp_rc)
    session.handle_pubcomp(ack)
    if comp_rc.is_error():
        with pytest.raises(MQTTError) as excinfo:
            handle.wait_for_ack(0.001)
        assert excinfo.value.reason_code == comp_rc
    else:
        assert handle.wait_for_ack(0.001) is ack


def test_session_publish_qos2_rec_error(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q2)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q2,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    session.handle_pubrec(MQTTPubRecPacket(reason_code=MQTTReasonCode.UnspecifiedError, packet_id=1))
    with pytest.raises(MQTTError) as excinfo:
        handle.wait_for_ack(0.001)
    assert excinfo.value.reason_code == MQTTReasonCode.UnspecifiedError

    mock_connection.send.assert_not_called()


def test_session_publish_qos2_disconnect_race(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = True

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q2)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        qos=MQTTQoS.Q2,
        packet_id=1,
    ))
    mock_connection.send.reset_mock()
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    mock_connection.can_send.return_value = True
    mock_connection.send.side_effect = InvalidStateError("TEST")

    session.handle_pubrec(MQTTPubRecPacket(packet_id=1))
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)

    mock_connection.send.assert_called_with(MQTTPubRelPacket(packet_id=1))
    mock_connection.send.reset_mock()


def test_session_publish_qos2_unconnected(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(ReceiveMaximum=255)))
    mock_connection.can_send.return_value = False
    mock_connection.send.side_effect = InvalidStateError("TEST")

    handle = session.publish("test/topic", b"test payload", qos=MQTTQoS.Q2)
    mock_connection.send.assert_not_called()
    with pytest.raises(TimeoutError):
        handle.wait_for_ack(0.001)


@pytest.mark.parametrize("db_path", [":memory:", ""])
@pytest.mark.parametrize("qos", [MQTTQoS.Q0, MQTTQoS.Q1, MQTTQoS.Q2])
def test_session_publish_alias(db_path: str, qos: MQTTQoS, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection, db_path=db_path)
    session.set_params(ConnectParams(client_id="test_client", clean_start=True))
    mock_connection.can_send.return_value = True

    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(TopicAliasMaximum=255)))

    session.publish("test/topic1", b"test payload", qos=qos, alias_policy=AliasPolicy.NEVER)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic1",
        payload=b"test payload",
        qos=qos,
        packet_id=1 if qos > 0 else 0,
    ))
    mock_connection.send.reset_mock()

    # A failed write before an alias is established should not increment the topic alias.
    mock_connection.send.side_effect = InvalidStateError("TEST")
    session.publish("test/topic2", b"test payload", qos=qos, alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic2",
        payload=b"test payload",
        qos=qos,
        packet_id=2 if qos > 0 else 0,
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    mock_connection.send.side_effect = None
    session.publish("test/topic3", b"test payload", qos=qos, alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="test/topic3",
        payload=b"test payload",
        qos=qos,
        packet_id=3 if qos > 0 else 0,
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    # A failed write after an alias is established should not affect aliases.
    mock_connection.send.side_effect = InvalidStateError("TEST")
    session.publish("test/topic3", b"test payload", qos=qos, alias_policy=AliasPolicy.TRY)
    mock_connection.send.assert_called_with(MQTTPublishPacket(
        topic="",
        payload=b"test payload",
        qos=qos,
        packet_id=4 if qos > 0 else 0,
        properties=MQTTPublishProps(TopicAlias=1),
    ))
    mock_connection.send.reset_mock()

    if qos > 0:
        with pytest.raises(ValueError):
            session.publish("test/topic4", b"test payload", qos=qos, alias_policy=AliasPolicy.ALWAYS)
    else:
        session.publish("test/topic4", b"test payload", qos=qos, alias_policy=AliasPolicy.ALWAYS)
        mock_connection.send.assert_called_with(MQTTPublishPacket(
            topic="test/topic4",
            payload=b"test payload",
            qos=qos,
            properties=MQTTPublishProps(TopicAlias=2),
        ))
        mock_connection.send.reset_mock()


def test_session_handle_publish_qos0(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
    )
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)


def test_session_handle_publish_qos1(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    mock_connection.can_send.return_value = True

    packet = MQTTPublishPacket(
        topic="test/topic",
        payload=b"test payload",
        packet_id=3,
        qos=MQTTQoS.Q1,
    )
    session.handle_publish(packet)
    mock_subscriptions.handle_publish.assert_called_once_with(packet)
    mock_connection.send.assert_called_once_with(MQTTPubAckPacket(packet_id=3))


@pytest.mark.parametrize("ack_rc", [MQTTReasonCode.Success, MQTTReasonCode.MessageRateTooHigh])
def test_session_handle_publish_qos2(ack_rc: MQTTReasonCode, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
     session = Session(mock_handlers, mock_subscriptions, mock_connection)
     mock_connection.can_send.return_value = True

     packet = MQTTPublishPacket(
         topic="test/topic",
         payload=b"test payload",
         packet_id=3,
         qos=MQTTQoS.Q2,
     )
     session.handle_publish(packet)
     mock_subscriptions.handle_publish.assert_called_once_with(packet)
     mock_subscriptions.handle_publish.reset_mock()
     mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))
     mock_connection.send.reset_mock()

     # Filter duplicates.
     session.handle_publish(packet)
     mock_subscriptions.handle_publish.assert_not_called()
     mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))
     mock_connection.send.reset_mock()

     session.handle_pubrel(MQTTPubRelPacket(reason_code=ack_rc, packet_id=3))
     mock_connection.send.assert_called_once_with(MQTTPubCompPacket(packet_id=3))
     mock_connection.send.reset_mock()
     # After PUBREL, message with same packet_id should be treated as a new application message.
     session.handle_publish(packet)
     mock_subscriptions.handle_publish.assert_called_once_with(packet)
     mock_connection.send.assert_called_once_with(MQTTPubRecPacket(packet_id=3))


def test_session_handle_connack_client_id(mocker: MockerFixture, mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    session.set_params(ConnectParams(clean_start=True))
    mock_persistence = mocker.Mock()
    session.persistence = mock_persistence
    mock_persistence.get.return_value = []
    session.handle_connack(MQTTConnAckPacket(properties=MQTTConnAckProps(AssignedClientIdentifier="foo")))
    assert mock_persistence.open.call_args[0][0] == "foo"


def test_session_handle_connack_error(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    with pytest.raises(MQTTError) as excinfo:
        session.handle_connack(MQTTConnAckPacket(reason_code=MQTTReasonCode.Banned))
    assert excinfo.value.reason_code == MQTTReasonCode.Banned


def test_session_handle_connack_no_client_id(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    with pytest.raises(MQTTError) as excinfo:
        session.handle_connack(MQTTConnAckPacket())
    assert excinfo.value.reason_code == MQTTReasonCode.ProtocolError


def test_session_slots(mock_handlers: Mock, mock_subscriptions: Mock, mock_connection: Mock) -> None:
    session = Session(mock_handlers, mock_subscriptions, mock_connection)
    assert not hasattr(session, "__dict__")
    assert all(hasattr(session, attr) for attr in session.__slots__), \
        [attr for attr in session.__slots__ if not hasattr(session, attr)]



"""PUBLISH, PUBACK, PUBREC, PUBREL, and PUBCOMP packets."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import ClassVar, Final, Mapping, TypeAlias

from .base import MQTTPacket
from ..error import MQTTError
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode, MQTTQoS
from ..property import (
    MQTTProperties,
    MQTTPublishProps,
    MQTTPubAckProps,
    MQTTPubRecProps,
    MQTTPubRelProps,
    MQTTPubCompProps,
)
from ..serialization import (
    encode_string,
    encode_varint,
    encode_uint16,
    decode_string,
    decode_uint8,
    decode_uint16,
)


HEAD_PUBLISH: Final = MQTTPacketType.PUBLISH << 4
HEAD_PUBACKS: Final[Mapping[MQTTPacketType, int]] = {
    MQTTPacketType.PUBACK: MQTTPacketType.PUBACK << 4,
    MQTTPacketType.PUBREC: MQTTPacketType.PUBREC << 4,
    MQTTPacketType.PUBREL: (MQTTPacketType.PUBREL << 4) + 0x02,
    MQTTPacketType.PUBCOMP: MQTTPacketType.PUBCOMP << 4,
}
FLAGS_PUBACKS: Final[Mapping[MQTTPacketType, int]] = {
    MQTTPacketType.PUBACK: 0,
    MQTTPacketType.PUBREC: 0,
    MQTTPacketType.PUBREL: 2,
    MQTTPacketType.PUBCOMP: 0,
}


@dataclass(match_args=True, slots=True)
class MQTTPublishPacket(MQTTPacket):
    packet_type = MQTTPacketType.PUBLISH
    props_type = MQTTPublishProps
    topic: str = ""
    payload: bytes = b""
    qos: MQTTQoS = MQTTQoS.Q0
    retain: bool = False
    packet_id: int = 0
    properties: MQTTPublishProps = field(default_factory=MQTTPublishProps)
    dup: bool = False

    def __str__(self) -> str:
        attrs = [
            f"topic={self.topic}",
            f"payload={len(self.payload)}B",
            f"qos={self.qos.value}",
            f"packet_id={self.packet_id}",
            f"retain={self.retain}",
            f"dup={self.dup}",
            f"properties={self.properties}",
        ]
        return f"PUBLISH[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        encoded = bytearray(encode_string(self.topic))
        if self.qos > MQTTQoS.Q0:
            encoded.extend(encode_uint16(self.packet_id))
        if self.properties:
            encoded.extend(self.properties.encode())
        else:
            encoded.append(0)
        encoded.extend(self.payload)
        head = HEAD_PUBLISH | self.retain | (self.qos.value << 1) | (self.dup << 3)
        encoded[0:0] = head.to_bytes(1, "big") + encode_varint(len(encoded))
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPublishPacket:
        try:
            qos = MQTTQoS((flags >> 1) & 0x03)
        except ValueError:
            raise MQTTError("Invalid QoS value", MQTTReasonCode.MalformedPacket)
        retain = (flags & 0x01) == 1
        dup = (flags & 0x08) == 8

        topic, topic_length = decode_string(data)
        offset = topic_length
        if qos > 0:
            packet_id = decode_uint16(data[offset:])[0]
            offset += 2
        else:
            packet_id = 0
        props, props_length = MQTTPublishProps.decode(data[offset:])
        offset += props_length
        payload = data[offset:].tobytes("A")
        return MQTTPublishPacket(
            topic,
            payload,
            qos=qos,
            retain=retain,
            dup=dup,
            packet_id=packet_id,
            properties=props,
        )


def _encode_pubacklike(packet: PubAckLikeT) -> bytes:
    """Encode a PUBACK-like packet."""
    head = HEAD_PUBACKS[packet.packet_type]
    encoded = bytearray(encode_uint16(packet.packet_id))
    if packet.reason_code != MQTTReasonCode.Success:
        encoded.append(packet.reason_code.value)
    if packet.properties:
        encoded.extend(packet.properties.encode())
    encoded[0:0] = encode_varint(len(encoded))
    encoded.insert(0, head)
    return bytes(encoded)


def _decode_pubacklike(cls: type[PubAckLikeT], flags: int, data: memoryview) -> tuple[int, int, PubAckLikePropsT]:
    """Decode a PUBACK, PUBREC, PUBREL, or PUBCOMP packet."""
    if flags != FLAGS_PUBACKS[cls.packet_type]:
        raise MQTTError(f"Invalid flags, expected {FLAGS_PUBACKS[cls.packet_type]} but got {flags}", MQTTReasonCode.MalformedPacket)
    # We will be kludging some types to make this work.

    offset = 0
    packet_id, packet_id_length = decode_uint16(data[offset:])
    offset += packet_id_length
    if offset == len(data):
        # Reason code and properties are optional.
        return cls(
            packet_id,
            MQTTReasonCode.Success,
            cls.props_type(),  # type: ignore
        )
    reason_code = MQTTReasonCode(decode_uint8(data[offset:])[0])
    offset += 1
    if offset == len(data):
        # Properties alone may be omitted.
        return cls(
            packet_id,
            reason_code,
            cls.props_type(),  # type: ignore
        )
    props, _ = cls.props_type.decode(data[offset:])
    return cls(
        packet_id,
        reason_code,
        props,  # type: ignore
    )


@dataclass(match_args=True, slots=True)
class MQTTPubAckPacket(MQTTPacket):
    packet_type = MQTTPacketType.PUBACK
    props_type: ClassVar[type[MQTTProperties]] = MQTTPubAckProps
    packet_id: int
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTPubAckProps = field(default_factory=MQTTPubAckProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"PUBACK[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        return _encode_pubacklike(self)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPubAckPacket:
        return _decode_pubacklike(cls, flags, data)  # type: ignore


@dataclass(match_args=True, slots=True)
class MQTTPubRecPacket(MQTTPacket):
    packet_type = MQTTPacketType.PUBREC
    props_type = MQTTPubRecProps
    packet_id: int
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTPubRecProps = field(default_factory=MQTTPubRecProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"PUBREC[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        return _encode_pubacklike(self)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPubRecPacket:
        return _decode_pubacklike(cls, flags, data)  # type: ignore


@dataclass(match_args=True, slots=True)
class MQTTPubRelPacket(MQTTPacket):
    packet_type = MQTTPacketType.PUBREL
    props_type = MQTTPubRelProps
    packet_id: int
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTPubRelProps = field(default_factory=MQTTPubRelProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"PUBREL[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        return _encode_pubacklike(self)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPubRelPacket:
        return _decode_pubacklike(cls, flags, data)  # type: ignore


@dataclass(match_args=True, slots=True)
class MQTTPubCompPacket(MQTTPacket):
    packet_type = MQTTPacketType.PUBCOMP
    props_type = MQTTPubCompProps
    packet_id: int
    reason_code: MQTTReasonCode = MQTTReasonCode.Success
    properties: MQTTPubCompProps = field(default_factory=MQTTPubCompProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_code={hex(self.reason_code)}",
            f"properties={self.properties}",
        ]
        return f"PUBCOMP[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        return _encode_pubacklike(self)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTPubCompPacket:
        return _decode_pubacklike(cls, flags, data)  # type: ignore


PubAckLikeT: TypeAlias = MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubRelPacket | MQTTPubCompPacket
PubAckLikePropsT: TypeAlias = MQTTPubAckProps | MQTTPubRecProps | MQTTPubRelProps | MQTTPubCompProps



from __future__ import annotations

import logging
from pathlib import Path
import pytest
import socket
import ssl
import tempfile
import threading
from typing import Any, Callable, Generator, TypeVar
import yaml

from tests.util.selfsigned import generate_selfsigned_cert

from ohmqtt.platform import HAS_AF_UNIX

logger = logging.getLogger(__name__)

LoopSelfT = TypeVar("LoopSelfT", bound="LoopbackSocket")


@pytest.fixture(scope="module")
def test_data_file(request: pytest.FixtureRequest) -> dict[str, Any]:
    """Load test data from a YAML file.

    The YAML file must be named after the test suite, and contain a mapping of test names to test data."""
    suite_name = request.module.__name__.split(".")[-1]
    data_path = Path("tests") / "data" / f"{suite_name}.yml"
    with data_path.open(encoding="utf-8") as f:
        return yaml.safe_load(f)  # type: ignore[no-any-return]


@pytest.fixture
def test_data(test_data_file: dict[str, Any], request: pytest.FixtureRequest) -> list[dict[str, Any]]:
    """Load test data for a specific test."""
    test_name = request.node.name
    return test_data_file[test_name]  # type: ignore[no-any-return]


class LoopbackSocket:
    """A pair of connected sockets for testing.

    Return an instance of this class from a mock to use as a socket in tests."""
    connect_calls: list[tuple[Any, ...] | str]

    def __init__(self) -> None:
        self.reset()
        self.test_close = self.testsock.close
        self.test_sendall = self.testsock.sendall
        self.test_recv = self.testsock.recv
        self.test_shutdown = self.testsock.shutdown
        self.close = self.mocksock.close
        self.detach = self.mocksock.detach
        self.family = self.mocksock.family
        self.fileno = self.mocksock.fileno
        self.getblocking = self.mocksock.getblocking
        self.getsockopt = self.mocksock.getsockopt
        self.gettimeout = self.mocksock.gettimeout
        self.proto = self.mocksock.proto
        self.recv = self.mocksock.recv
        self.recv_into = self.mocksock.recv_into
        self.send = self.mocksock.send
        self.sendall = self.mocksock.sendall
        self.setblocking = self.mocksock.setblocking
        self.shutdown = self.mocksock.shutdown
        self.type = self.mocksock.type

    def __enter__(self: LoopSelfT) -> LoopSelfT:
        return self

    def __exit__(self, *args: Any) -> None:
        self.mocksock.close()
        self.testsock.close()

    def __del__(self) -> None:
        self.mocksock.close()
        self.testsock.close()

    def reset(self) -> None:
        """Reset the socket pair."""
        self.mocksock, self.testsock = socket.socketpair()
        self.testsock.settimeout(5.0)
        self.connect_calls = []

    def connect(self, address: tuple[Any, ...] | str) -> None:
        self.connect_calls.append(address)

    def setsockopt(self, level: int, optname: int, value: int) -> None:
        if optname == socket.TCP_NODELAY:
            # Where AF_UNIX exists, do not set TCP_NODELAY on either side of the socket.
            # We are spoofing TCP but the socketpair is a Unix domain socket.
            if HAS_AF_UNIX:
                logger.info("Not setting TCP_NODELAY on Unix domain socket")
                return
        self.mocksock.setsockopt(level, optname, value)


@pytest.fixture
def loopback_socket() -> Generator[LoopbackSocket, None, None]:
    with LoopbackSocket() as loop:
        yield loop


class LoopbackTLSSocket(LoopbackSocket):
    """A pair of connected sockets for testing. The test side is wrapped in an SSL context.

    Return an instance of this class from a mock to use as a socket in tests.

    You must call test_do_handshake() before using either end of the socket."""
    testsock: ssl.SSLSocket

    def __init__(self) -> None:
        super().__init__()
        self._wrap_socket()

    def reset(self) -> None:
        """Reset the socket pair."""
        super().reset()
        self._wrap_socket()

    def _wrap_socket(self) -> None:
        """Wrap the test side of the socket in an SSL context."""
        self.server_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
        self.server_context.minimum_version = ssl.TLSVersion.TLSv1_3
        self.server_context.maximum_version = ssl.TLSVersion.TLSv1_3
        self.cert_pem, key_pem = generate_selfsigned_cert("localhost")
        with tempfile.TemporaryDirectory() as tmpdir:
            certfile = Path(tmpdir) / "cert.pem"
            keyfile = Path(tmpdir) / "key.pem"
            with certfile.open("wb") as f:
                f.write(self.cert_pem)
            with keyfile.open("wb") as f:
                f.write(key_pem)
            self.server_context.load_cert_chain(certfile, keyfile)
        self.testsock = self.server_context.wrap_socket(self.testsock, server_side=True, do_handshake_on_connect=False)

    def _do_handshake(self) -> None:
        self.testsock.do_handshake()

    def test_do_handshake(self) -> None:
        """Call do_handshake() on the test side of the socket in a thread, to avoid blocking the test."""
        self.handshake_thread = threading.Thread(target=self._do_handshake, daemon=True)
        self.handshake_thread.start()


@pytest.fixture
def loopback_tls_socket() -> Generator[LoopbackTLSSocket, None, None]:
    with LoopbackTLSSocket() as loop:
        yield loop


@pytest.fixture
def ssl_client_context() -> Callable[[bytes], ssl.SSLContext]:
    """Provides a function for getting a new SSL client context with provided certificate."""
    def _ssl_client_context(cert_pem: bytes) -> ssl.SSLContext:
        context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        with tempfile.TemporaryDirectory() as tmpdir:
            certfile = Path(tmpdir) / "cert.pem"
            with certfile.open("wb") as f:
                f.write(cert_pem)
            context.load_verify_locations(certfile)
        return context
    return _ssl_client_context



from .mqtt_spec import MQTTReasonCode


class MQTTError(Exception):
    def __init__(self, message: str, reason_code: MQTTReasonCode = MQTTReasonCode.UnspecifiedError) -> None:
        super().__init__(message)
        self.reason_code = reason_code

    def __str__(self) -> str:
        return f"{super().__str__()} (reason code: {MQTTReasonCode(self.reason_code).name}{{{hex(self.reason_code)}}})"



#!/usr/bin/env python3
"""This example demonstrates a simple RPC caller.

It sends an RPC request to a specific topic and waits for the response.

The ResponseTopic property is used to specify the topic to which the response should be sent."""

import argparse
import threading
import uuid
from typing import Callable

from ohmqtt import Client, MQTTPublishPacket, MQTTPublishProps, MQTTQoS


RPCCallback = Callable[[bytes], None]


class RPCCaller:
    """A simple RPC caller."""
    def __init__(self, client: Client) -> None:
        self.client = client
        self.callbacks: dict[str, RPCCallback] = {}

    def send_request(self, payload: bytes, callback: RPCCallback) -> None:
        # Expect the response to be sent to a unique topic.
        unique_id = str(uuid.uuid4())
        response_topic = f"ohmqtt/examples/rpc/response/{unique_id}"

        # Store the callback for this request.
        self.callbacks[unique_id] = callback

        # Subscribe to the response topic.
        self.client.subscribe(response_topic, self.handle_response)

        # Publish the request with the necessary properties.
        self.client.publish(
            "ohmqtt/examples/rpc/request",
            payload,
            qos=MQTTQoS.Q2,
            properties=MQTTPublishProps(ResponseTopic=response_topic),
        ).wait_for_ack(timeout=5.0)

        print(f"Sent RPC request with response topic: {response_topic}")

    def handle_response(self, client: Client, msg: MQTTPublishPacket) -> None:
        """Handle incoming RPC responses."""
        try:
            callback = self.callbacks.pop(msg.topic.split("/")[-1])
        except KeyError:
            print("Couldn't find callback for response")
            return
        else:
            # Call the callback with the response payload.
            callback(msg.payload)
        finally:
            # Unsubscribe from the response topic.
            client.unsubscribe(msg.topic)


def main(args: argparse.Namespace) -> None:
    with Client() as client:
        rpc_caller = RPCCaller(client)
        client.connect(args.address)
        print("*** Waiting for connection...")
        client.wait_for_connect(timeout=5.0)

        # Set up a callback which helps us block until a response is received.
        recvd = threading.Event()
        def callback(payload: bytes) -> None:
            print(f"*** Received RPC response: {payload.decode()}")
            recvd.set()  # Signal that the response has been received.

        while True:
            recvd.clear()  # Reset the event for the next request.
            try:
                payload = input("*** Enter something fun (or 'exit' to quit): ")
                if payload.lower() == "exit":
                    break
                rpc_caller.send_request(payload.encode(), callback)
                recvd.wait(timeout=5.0)  # Wait for the response to be received.
            except (EOFError, KeyboardInterrupt):
                print("\n*** Shutting down RPC caller...")
                break


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)



from __future__ import annotations

from typing import Any, cast, Callable, TypeVar

from .types import ReceivablePacketT
from ..packet import (
    MQTTPacket,
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
    MQTTSubAckPacket,
    MQTTUnsubAckPacket,
    MQTTConnAckPacket,
    MQTTAuthPacket,
    MQTTDisconnectPacket,
)
PacketT = TypeVar("PacketT", bound=ReceivablePacketT)


class MessageHandlers:
    """Container for intentional registration of message handlers for MQTT packets.

    Handlers may only be registered within the context manager.

    The context manager may only be entered once."""
    __slots__ = ("_handlers", "_registered", "_registering")

    def __init__(self) -> None:
        # Typing is kludged internally for brevity and speed.  But mostly speed.
        self._handlers: dict[type[MQTTPacket], list[Callable[[Any], None]]] = {
            MQTTPublishPacket: [],
            MQTTPubAckPacket: [],
            MQTTPubRecPacket: [],
            MQTTPubRelPacket: [],
            MQTTPubCompPacket: [],
            MQTTSubAckPacket: [],
            MQTTUnsubAckPacket: [],
            MQTTConnAckPacket: [],
            MQTTAuthPacket: [],
            MQTTDisconnectPacket: [],
        }
        self._registering = False
        self._registered = False

    def __enter__(self) -> MessageHandlers:
        """Start registering the message handlers.

        :raises RuntimeError: Message handlers are already registered (the context manager may only be entered once)."""
        if self._registered:
            raise RuntimeError("Message handlers already registered")
        self._registering = True
        return self

    def __exit__(self, *args: object) -> None:
        """Stop registering the message handlers."""
        self._registering = False
        self._registered = True

    def get_handlers(self, packet_type: type[PacketT]) -> list[Callable[[PacketT], None]]:
        """Get the handlers for a given packet type.

        :raises RuntimeError: Message handlers registration has not been completed."""
        if not self._registered:
            raise RuntimeError("Message handlers not registered")
        return cast(list[Callable[[PacketT], None]], self._handlers[packet_type])

    def register(self, packet_type: type[PacketT], handler: Callable[[PacketT], None]) -> None:
        """Register a handler for a given packet type.

        :raises RuntimeError: Not in the context manager."""
        if not self._registering:
            raise RuntimeError("Message handlers not in registration block")
        self._handlers[packet_type].append(handler)

    def handle(self, packet: ReceivablePacketT) -> list[Exception]:
        """Handle a packet by calling the appropriate handlers.

        Guarantees that all handlers are called in the order they were registered.

        Guarantees that all handlers are run regardless of Exceptions.

        :return: A list of Exceptions raised by handlers."""
        if not self._registered:
            raise RuntimeError("Message handlers not registered")
        exceptions = []
        handlers = self.get_handlers(type(packet))
        for handler in handlers:
            try:
                handler(packet)
            except Exception as exc:  # noqa: PERF203
                exceptions.append(exc)
        return exceptions



from __future__ import annotations

from typing import Final

from .address import Address as Address
from .fsm import FSM
from .fsm import InvalidStateError as InvalidStateError
from .handlers import MessageHandlers as MessageHandlers
from .states import (
    ClosingState,
    ConnectedState,
    ConnectingState,
    ReconnectWaitState,
    ShutdownState,
    ClosedState,
)
from .types import ConnectParams as ConnectParams
from .types import StateEnvironment, ReceivablePacketT, SendablePacketT
from ..error import MQTTError
from ..logger import get_logger

logger: Final = get_logger("connection")


class Connection:
    """Interface for the MQTT connection."""
    __slots__ = ("_handlers", "fsm")

    def __init__(self, handlers: MessageHandlers) -> None:
        state_env = StateEnvironment(packet_callback=self.handle_packet)
        self.fsm = FSM(env=state_env, init_state=ClosedState, error_state=ShutdownState)
        self._handlers = handlers

    def handle_packet(self, packet: ReceivablePacketT) -> None:
        """Handle incoming packets by routing them to registered handlers."""
        logger.debug("<--- %s", packet)
        exceptions = self._handlers.handle(packet)
        if exceptions:
            if any(True for exc in exceptions if isinstance(exc, MQTTError)):
                # If there is an MQTTError, raise it.
                raise next(exc for exc in exceptions if isinstance(exc, MQTTError))
            # Otherwise. raise the first exception.
            raise exceptions[0]

    def can_send(self) -> bool:
        """Check if the connection is in a state where data can be sent.

        :return: True if the connection is in a state where data can be sent, False otherwise."""
        with self.fsm.lock:
            state = self.fsm.get_state()
            return state is ConnectedState

    def send(self, packet: SendablePacketT) -> None:
        """Send data to the connection.

        :raises InvalidStateError: If the connection is not in a state where data can be sent."""
        with self.fsm.lock:
            if not self.can_send():
                state = self.fsm.get_state()
                raise InvalidStateError(f"Cannot send data in state {state.__name__}")
            logger.debug("---> %s", packet)
            self.fsm.env.packet_buffer.append(packet)
            self.fsm.selector.interrupt()

    def connect(self, params: ConnectParams) -> None:
        """Connect to the MQTT broker."""
        with self.fsm.lock:
            if not self.fsm.get_state().can_transition_to(ConnectingState):
                state = self.fsm.get_state()
                raise InvalidStateError(f"Cannot connect() from state {state.__name__}, disconnect() first")
            self.fsm.set_params(params)
            self.fsm.change_state(ConnectingState)
            self.fsm.selector.interrupt()

    def disconnect(self) -> None:
        """Disconnect from the MQTT broker."""
        with self.fsm.lock:
            self.fsm.request_state(ClosingState)
            self.fsm.selector.interrupt()

    def shutdown(self) -> None:
        """Shutdown the connection."""
        with self.fsm.lock:
            self.fsm.request_state(ShutdownState)
            self.fsm.selector.interrupt()

    def is_connected(self) -> bool:
        """Check if the connection is established."""
        return self.fsm.get_state() == ConnectedState

    def wait_for_connect(self, timeout: float | None = None) -> bool:
        """Wait for the connection to be established.

        :return: True if the connection is established, False if the timeout is reached."""
        return self.fsm.wait_for_state((ConnectedState,), timeout)

    def wait_for_disconnect(self, timeout: float | None = None) -> bool:
        """Wait for the connection to be closed.

        :return: True if the connection is closed, False if the timeout is reached."""
        return self.fsm.wait_for_state((ClosedState, ShutdownState, ReconnectWaitState), timeout)

    def wait_for_shutdown(self, timeout: float | None = None) -> bool:
        """Wait for the connection to be closed and finalized.

        :return: True if the connection is closed, False if the timeout is reached."""
        return self.fsm.wait_for_state((ShutdownState,), timeout)

    def loop_once(self, max_wait: float | None = 0.0) -> None:
        """Run a single iteration of the state machine.

        If max_wait is None, wait indefinitely. Otherwise, wait for the specified time."""
        self.fsm.loop_once(max_wait)

    def loop_forever(self) -> None:
        """Run the state machine until the connection is shut down."""
        self.fsm.loop_until_state((ShutdownState,))

    def loop_until_connected(self, timeout: float | None = None) -> bool:
        """Run the state machine until the connection is established.

        :return: True if the connection is established, False otherwise."""
        return self.fsm.loop_until_state((ConnectedState,), timeout)



from functools import partial
from pathlib import Path
import tempfile
from typing import Final, Generator

import pytest
from pytest_mock import MockerFixture

from ohmqtt.error import MQTTError
from ohmqtt.mqtt_spec import MQTTQoS, MQTTReasonCode
from ohmqtt.packet import (
    MQTTPublishPacket,
    MQTTPubAckPacket,
    MQTTPubRecPacket,
    MQTTPubRelPacket,
    MQTTPubCompPacket,
)
from ohmqtt.persistence.base import LostMessageError, Persistence
from ohmqtt.persistence.in_memory import InMemoryPersistence
from ohmqtt.persistence.sqlite import SQLitePersistence
from ohmqtt.property import MQTTPublishProps
from ohmqtt.topic_alias import AliasPolicy


SQLiteInMemory = partial(SQLitePersistence, ":memory:", db_fast=True)


@pytest.fixture
def tempdbpath() -> Generator[str, None, None]:
    """Fixture to create a temporary directory for testing."""
    with tempfile.TemporaryDirectory() as tempdir:
        yield str(Path(tempdir) / "test.db")


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_open(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:

        persistence.open("test_client")
        persistence.add(
            "test/topic",
            b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
            alias_policy=AliasPolicy.TRY,
        )
        assert len(persistence) == 1
        mids = persistence.get(1)
        rendered = persistence.render(mids[0])
        assert isinstance(rendered.packet, MQTTPublishPacket)
        assert not rendered.packet.dup

        persistence.open("test_client")
        assert len(persistence) == 1
        rendered.packet.dup = True
        mids = persistence.get(1)
        assert rendered.packet == persistence.render(mids[0]).packet

        persistence.open("test_client2")
        assert len(persistence) == 0


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_happy_path_qos1(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")
        assert len(persistence) == 0

        # Add a message to the store.
        handle = persistence.add(
            "test/topic",
            b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
            alias_policy=AliasPolicy.TRY,
        )
        assert len(persistence) == 1

        # Retrieve the PUBLISH from the store.
        message_ids = persistence.get(10)
        assert len(message_ids) == 1
        expected_packet = MQTTPublishPacket(
            packet_id=1,
            topic="test/topic",
            payload=b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
        )

        # Render the message, marking it as inflight.
        rendered = persistence.render(message_ids[0])
        assert rendered.packet == expected_packet
        assert rendered.alias_policy == AliasPolicy.TRY
        assert len(persistence) == 1

        # We should not be able to retrieve the message again.
        assert len(persistence.get(10)) == 0

        # Acknowledge the message.
        ack_packet = MQTTPubAckPacket(packet_id=1)
        persistence.ack(ack_packet)
        assert len(persistence) == 0
        assert handle.wait_for_ack(0.001) == ack_packet


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_happy_path_qos2(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")
        assert len(persistence) == 0

        # Add a message to the store.
        handle = persistence.add(
            "test/topic",
            b"test payload",
            qos=MQTTQoS.Q2,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
            alias_policy=AliasPolicy.TRY,
        )
        assert len(persistence) == 1

        # Retrieve the PUBLISH from the store.
        message_ids = persistence.get(10)
        assert len(message_ids) == 1
        expected_publish = MQTTPublishPacket(
            packet_id=1,
            topic="test/topic",
            payload=b"test payload",
            qos=MQTTQoS.Q2,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
        )

        # Render the message, marking it as inflight.
        rendered = persistence.render(message_ids[0])
        assert rendered.packet == expected_publish
        assert rendered.alias_policy == AliasPolicy.TRY
        assert len(persistence) == 1

        # We should not be able to retrieve the message again.
        assert len(persistence.get(10)) == 0

        # Acknowledge the message.
        persistence.ack(MQTTPubRecPacket(packet_id=1))
        assert len(persistence) == 1
        assert handle.ack is None

        # Retrieve the PUBREL from the store.
        message_ids = persistence.get(10)
        assert len(message_ids) == 1
        expected_pubrel = MQTTPubRelPacket(packet_id=1)

        # Render the message, marking it as inflight.
        rendered = persistence.render(message_ids[0])
        assert rendered.packet == expected_pubrel
        assert rendered.alias_policy == AliasPolicy.NEVER
        assert len(persistence) == 1

        # We should not be able to retrieve the message again.
        assert len(persistence.get(10)) == 0

        # Acknowledge the message.
        ack_packet = MQTTPubCompPacket(packet_id=1)
        persistence.ack(ack_packet)
        assert len(persistence) == 0
        assert handle.wait_for_ack(0.001) == ack_packet


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_properties(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        # Add a message with all the properties.
        packet = MQTTPublishPacket(
            packet_id=1,
            topic="test/topic",
            payload=b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(
                ResponseTopic="response/topic",
                CorrelationData=b"correlation data",
                MessageExpiryInterval=60,
                UserProperty=[("key", "value")],
                SubscriptionIdentifier={123},
            ),
        )
        persistence.add(
            topic=packet.topic,
            payload=packet.payload,
            qos=packet.qos,
            retain=packet.retain,
            properties=packet.properties,
            alias_policy=AliasPolicy.TRY,
        )
        assert len(persistence) == 1

        # Retrieve the PUBLISH from the store.
        message_ids = persistence.get(10)
        rendered = persistence.render(message_ids[0])
        assert rendered.packet == packet


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_loose_alias(persistence_class: type[Persistence]) -> None:
    """We should not be able to add a message with an alias policy of ALWAYS."""
    with persistence_class() as persistence:
        persistence.open("test_client")

        # Add a message with all the properties.
        packet = MQTTPublishPacket(
            packet_id=1,
            topic="test/topic",
            payload=b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(),
        )
        with pytest.raises(ValueError):
            persistence.add(
                topic=packet.topic,
                payload=packet.payload,
                qos=packet.qos,
                retain=packet.retain,
                properties=packet.properties,
                alias_policy=AliasPolicy.ALWAYS,
            )
        assert len(persistence) == 0


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_incoming_qos2(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        pub_packet = MQTTPublishPacket(
            packet_id=1,
            qos=MQTTQoS.Q1,
        )
        with pytest.raises(ValueError):
            persistence.check_rec(pub_packet)
        with pytest.raises(ValueError):
            persistence.set_rec(pub_packet)

        pub_packet.qos = MQTTQoS.Q2

        assert persistence.check_rec(pub_packet) is True
        persistence.set_rec(pub_packet)
        assert persistence.check_rec(pub_packet) is False

        rel_packet = MQTTPubRelPacket(packet_id=1)
        persistence.rel(rel_packet)

        assert persistence.check_rec(pub_packet) is True


@pytest.mark.parametrize("qos", [MQTTQoS.Q1, MQTTQoS.Q2])
@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_queue(qos: MQTTQoS, persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        rng = range(1, 3)
        for n in rng:
            persistence.add(
                topic=str(n),
                payload=b"bar",
                qos=qos,
                retain=False,
                properties=MQTTPublishProps(),
                alias_policy=AliasPolicy.NEVER,
            )
        for n in rng:
            queued = persistence.get(3)
            rendered = persistence.render(queued[0])
            assert isinstance(rendered.packet, MQTTPublishPacket)
            assert rendered.packet.topic == str(n)
            assert rendered.packet.packet_id == n


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_render_order(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        for _ in (1, 2):
            persistence.add(
                topic="foo",
                payload=b"bar",
                qos=MQTTQoS.Q1,
                retain=False,
                properties=MQTTPublishProps(),
                alias_policy=AliasPolicy.NEVER,
            )
        queued = persistence.get(3)
        with pytest.raises(ValueError):
            persistence.render(queued[1])


@pytest.mark.parametrize("persistence_cls", [SQLiteInMemory, InMemoryPersistence])
@pytest.mark.parametrize("ack_cls", [MQTTPubAckPacket, MQTTPubRecPacket, MQTTPubCompPacket])
def test_persistence_unknown_ack(persistence_cls: type[Persistence], ack_cls: type[MQTTPubAckPacket | MQTTPubRecPacket | MQTTPubCompPacket]) -> None:
    with persistence_cls() as persistence:
        persistence.open("test_client")

        ack_packet = ack_cls(packet_id=42)
        with pytest.raises(ValueError, match="Unknown packet_id: 42"):
            persistence.ack(ack_packet)


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_error_ack(persistence_class: type[Persistence]) -> None:
    """An error in PUBREC must complete a QoS2 message."""
    with persistence_class() as persistence:
        persistence.open("test_client")

        handle = persistence.add(
            "test/topic",
            b"test payload",
            qos=MQTTQoS.Q2,
            retain=False,
            properties=MQTTPublishProps(),
            alias_policy=AliasPolicy.NEVER,
        )
        message_ids = persistence.get(10)
        persistence.render(message_ids[0])
        rec_packet = MQTTPubRecPacket(packet_id=1, reason_code=MQTTReasonCode.UnspecifiedError)
        persistence.ack(rec_packet)

        assert handle.ack == rec_packet
        assert isinstance(handle.exc, MQTTError)
        assert handle.exc.reason_code == MQTTReasonCode.UnspecifiedError

        with pytest.raises(ValueError, match="Unknown packet_id: 1"):
            persistence.ack(MQTTPubCompPacket(packet_id=1))


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_lost_message(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        handle = persistence.add(
            "test/topic",
            b"test payload",
            qos=MQTTQoS.Q2,
            retain=False,
            properties=MQTTPublishProps(),
            alias_policy=AliasPolicy.NEVER,
        )
        message_ids = persistence.get(10)
        persistence.render(message_ids[0])
        persistence.clear()

        with pytest.raises(LostMessageError):
            handle.wait_for_ack(0.001)


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_unknown_render(persistence_class: type[Persistence]) -> None:
    with persistence_class() as persistence:
        persistence.open("test_client")

        with pytest.raises(KeyError):
            persistence.render(42)


@pytest.mark.parametrize("persistence_class", [SQLiteInMemory, InMemoryPersistence])
def test_persistence_unlimited_ids(persistence_class: type[Persistence]) -> None:
    "Should be able to queue unlimited messages."
    with persistence_class() as persistence:
        persistence.open("test_client")

        blocks: Final = 3  # How many blocks of 65535 packets to try
        num: Final = 65535 * blocks  # Total packets
        for _ in range(num):
            persistence.add("foo", b"bar", MQTTQoS.Q1, False, MQTTPublishProps(), AliasPolicy.NEVER)

        checked = 0
        for block in range(blocks):
            for n in range(1, 65536):
                checked += 1
                try:
                    pending = persistence.get(1)
                    rendered = persistence.render(pending[0])
                    assert isinstance(rendered.packet, MQTTPublishPacket)
                    assert rendered.packet.packet_id == n
                except Exception:
                    print(f"{block=} {n=}")
                    raise
                persistence.ack(MQTTPubAckPacket(packet_id=rendered.packet.packet_id))
        assert checked == num


@pytest.mark.parametrize("db_fast", [True, False])
def test_persistence_sqlite_open(db_fast: bool, tempdbpath: str) -> None:
    """Test the SQLitePersistence class with a resume scenario."""
    with SQLitePersistence(tempdbpath, db_fast=db_fast) as persistence:
        persistence.open("test_client")
        assert len(persistence) == 0

        # Add a message to the store.
        outgoing_packet = MQTTPublishPacket(
            topic="test/topic",
            payload=b"test payload",
            qos=MQTTQoS.Q1,
            retain=False,
            properties=MQTTPublishProps(ResponseTopic="response/topic"),
        )
        persistence.add(
            topic=outgoing_packet.topic,
            payload=outgoing_packet.payload,
            qos=outgoing_packet.qos,
            retain=outgoing_packet.retain,
            properties=outgoing_packet.properties,
            alias_policy=AliasPolicy.TRY,
        )
        assert len(persistence) == 1

        # Mark the message inflight by rendering it.
        persistence.render(1)
        assert len(persistence.get(1)) == 0

        # Simulate an incoming QoS 2 message.
        incoming_packet = MQTTPublishPacket(packet_id=1, qos=MQTTQoS.Q2)
        persistence.set_rec(incoming_packet)

        # Close and reopen the persistence store.
        persistence = SQLitePersistence(tempdbpath)
        # This should not clear the store.
        persistence.open("test_client")
        assert len(persistence) == 1

        # We should be able to retrieve the message again.
        message_ids = persistence.get(1)
        assert len(message_ids) == 1
        # When rendering a second time, the packet should have the dup flag set.
        outgoing_packet.dup = True
        outgoing_packet.packet_id = 1
        rendered = persistence.render(message_ids[0])
        assert rendered.packet == outgoing_packet
        assert rendered.alias_policy == AliasPolicy.TRY

        # We should filter a duplicate incoming QoS 2 message.
        assert persistence.check_rec(incoming_packet) is False

        # Now open with a different client ID.
        persistence = SQLitePersistence(tempdbpath)
        # This should clear the store.
        persistence.open("test_client_2")
        assert len(persistence) == 0
        assert persistence.check_rec(incoming_packet) is True


def test_persistence_sqlite_schema_version(mocker: MockerFixture) -> None:
    """Reject opening a database with the wrong schema version."""
    mocker.patch("ohmqtt.persistence.sqlite.SCHEMA_VERSION", 9999)
    with pytest.raises(Exception, match=r"Database version .* does not match library version .*"):
        SQLitePersistence(":memory:")



#!/usr/bin/env python3
"""This example demonstrates a single-threaded application which reacts to incoming messages.

Run this alongside the `examples/publish.py` example to see messages being received."""

import argparse

from ohmqtt import Client, MQTTPublishPacket


def main(args: argparse.Namespace) -> None:
    client = Client()
    client.connect(args.address)
    client.loop_until_connected(timeout=5.0)

    def callback(client: Client, packet: MQTTPublishPacket) -> None:
        print(f"Received message: {packet.payload.decode()}")

    client.subscribe("ohmqtt/examples/publish", callback=callback)
    try:
        client.loop_forever()
    except KeyboardInterrupt:
        client.shutdown()
        client.loop_once(max_wait=0)


if __name__ == "__main__":
    from .args import parser
    args = parser.parse_args()
    main(args)



from __future__ import annotations

import ssl
from typing import Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .mqtt_handshake_connack import MQTTHandshakeConnAckState
from ..types import ConnectParams, StateData, StateEnvironment
from ..wslib import OpCode, frame_ws_data
from ...logger import get_logger
from ...packet import MQTTConnectPacket

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.mqtt_handshake_connect")


class MQTTHandshakeConnectState(FSMState):
    """Sending MQTT CONNECT packet to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        connect_packet = MQTTConnectPacket(
            client_id=params.client_id,
            protocol_version=params.protocol_version,
            clean_start=params.clean_start,
            keep_alive=params.keepalive_interval,
            properties=params.connect_properties,
            will_topic=params.will_topic,
            will_payload=params.will_payload,
            will_qos=params.will_qos,
            will_retain=params.will_retain,
            will_props=params.will_properties,
            username=params.username,
            password=params.password,
        )
        logger.debug("---> %s", connect_packet)
        state_data.write_buffer.clear()
        payload = connect_packet.encode()
        if params.address.is_websocket():
            ws_frame = frame_ws_data(OpCode.BINARY, payload)
            state_data.write_buffer.extend(ws_frame)
        else:
            state_data.write_buffer.extend(payload)

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("MQTT CONNECT keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            num_sent = state_data.sock.send(state_data.write_buffer)
            if num_sent == 0:
                logger.error("MQTT CONNECT send returned 0 bytes, closing connection")
                fsm.change_state(ClosedState)
                return True
            if num_sent < len(state_data.write_buffer):
                # Not all data was sent, wait for writable again.
                logger.debug("Not all CONNECT data was sent, waiting for writable again: wrote: %d", num_sent)
                del state_data.write_buffer[:num_sent]
                return False
            state_data.write_buffer.clear()
            fsm.change_state(MQTTHandshakeConnAckState)
            return True
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            # The write was blocked, wait for the socket to be writable.
            if max_wait is None or max_wait > 0.0:
                with fsm.selector:
                    timeout = state_data.timeout.get_timeout(max_wait)
                    fsm.selector.select(write=True, timeout=timeout)
        return False



import pytest

from ohmqtt.connection.handlers import MessageHandlers
from ohmqtt.packet import MQTTPublishPacket


def test_handlers() -> None:
    handlers = MessageHandlers()
    calls = []
    def callback(packet: MQTTPublishPacket) -> None:
        calls.append(packet)
    class TestError(Exception):
        pass
    def broken_callback(packet: MQTTPublishPacket) -> None:
        raise TestError("This handler is broken")

    with pytest.raises(RuntimeError):
        handlers.get_handlers(MQTTPublishPacket)
    with pytest.raises(RuntimeError):
        handlers.register(MQTTPublishPacket, callback)
    with pytest.raises(RuntimeError):
        handlers.handle(MQTTPublishPacket())

    with handlers:
        handlers.register(MQTTPublishPacket, broken_callback)
        handlers.register(MQTTPublishPacket, callback)
    assert handlers.get_handlers(MQTTPublishPacket) == [broken_callback, callback]
    packet = MQTTPublishPacket(topic="test/topic", payload=b"test payload")
    exc_list = handlers.handle(packet)
    assert len(exc_list) == 1
    assert isinstance(exc_list[0], TestError)
    assert calls == [packet]

    with pytest.raises(RuntimeError):
        with handlers:
            pass



from __future__ import annotations

import socket
from typing import Final, TYPE_CHECKING

from .base import FSMState
from ..types import ConnectParams, StateData, StateEnvironment
from ..wslib import frame_ws_data, OpCode
from ...logger import get_logger
from ...packet import MQTTDisconnectPacket

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.closed")


class ClosedState(FSMState):
    """Connection is closed.

    All buffers will be flushed and the socket will be closed after conditionally sending a DISCONNECT."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        # Avoid circular imports.
        from .connecting import ConnectingState  # noqa: PLC0415
        from .reconnect_wait import ReconnectWaitState  # noqa: PLC0415

        if state_data.open_called:
            state_data.open_called = False
            if state_data.disconnect_rc is not None and not state_data.write_buffer:
                disconnect_packet = MQTTDisconnectPacket(reason_code=state_data.disconnect_rc)
                # Try to send a DISCONNECT packet, but no problem if we can't.
                payload = disconnect_packet.encode()
                if params.address.is_websocket():
                    payload = frame_ws_data(OpCode.BINARY, payload)
                try:
                    state_data.sock.send(payload)
                    logger.debug("---> %s", disconnect_packet)
                except OSError:
                    logger.debug("Failed to send DISCONNECT packet")
            if params.reconnect_delay > 0 and fsm.requested_state == ConnectingState:
                fsm.change_state(ReconnectWaitState)
        try:
            state_data.sock.shutdown(socket.SHUT_RDWR)
        except OSError as exc:
            logger.debug("Error while shutting down socket: %s", exc)
        try:
            state_data.sock.close()
        except OSError as exc:
            logger.debug("Error while closing socket: %s", exc)
        state_data.decoder.reset()
        state_data.ws_decoder.reset()
        state_data.ws_handshake_buffer.clear()
        state_data.ws_nonce = ""
        state_data.write_buffer.clear()
        env.packet_buffer.clear()



from dataclasses import dataclass, field
import socket
import ssl
from typing import NamedTuple, Final

from ..error import MQTTError
from ..logger import get_logger
from ..mqtt_spec import MQTTReasonCode
from ..packet import decode_packet_from_parts, MQTTPacket

logger: Final = get_logger("connection.decoder")


class VarintDecodeResult(NamedTuple):
    """Result of decoding a variable length integer, in part or whole.

    This state can be used to resume decoding if the socket doesn't have enough data."""
    value: int
    multiplier: int
    complete: bool


InitVarintDecodeState: Final = VarintDecodeResult(0, 1, False)


class ClosedSocketError(Exception):
    """Exception raised when the socket is closed."""


class WantReadError(Exception):
    """Indicates that the socket is not ready for reading."""


@dataclass(slots=True)
class IncrementalDecoder:
    """Incremental decoder for MQTT messages coming in from a socket."""
    head: int = field(default=-1, init=False)  #: The first byte of the packet.
    length: VarintDecodeResult = field(default=InitVarintDecodeState, init=False)  #: Variable integer decoding state of the packet length.
    data: bytearray = field(init=False, default_factory=bytearray)  #: The remaining data of the packet.

    def reset(self) -> None:
        """Reset the decoder state."""
        self.head = -1
        self.length = InitVarintDecodeState
        self.data.clear()

    def _recv_one_byte(self, sock: socket.socket | ssl.SSLSocket) -> int:
        """Receive one byte from the socket.

        :raises WantReadError: The socket is not ready for reading.
        :raises ClosedSocketError: The socket is closed."""
        try:
            data = sock.recv(1)
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            raise WantReadError("Socket not ready for reading")
        if not data:
            raise ClosedSocketError("Socket closed")
        return data[0]

    def _extract_head(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Extract the head byte of the next packet from the socket, if needed."""
        if self.head != -1:
            return
        self.head = self._recv_one_byte(sock)

    def _extract_length(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Incrementally decode a variable length integer from a socket, if needed.

        :raises WantReadError: The socket is not ready for reading."""
        # See ohmqtt.serialization.decode_varint for a cleaner implementation.
        assert self.head != -1, "Can't extract length without head"
        if self.length.complete:
            return
        result = self.length.value
        mult = self.length.multiplier
        try:
            while mult < 0x200000:  # This magic is the mult value after pulling 4 bytes.
                byte = self._recv_one_byte(sock)
                result += byte % 0x80 * mult
                if byte < 0x80:
                    # We have the complete varint.
                    self.length = VarintDecodeResult(result, mult, True)
                    return
                mult *= 0x80
            raise MQTTError("Varint overflow", MQTTReasonCode.MalformedPacket)
        except WantReadError:
            # Not done yet, the socket is neither closed nor ready for reading.
            # Save the partial state and re-raise.
            self.length = VarintDecodeResult(result, mult, False)
            raise

    def _extract_data(self, sock: socket.socket | ssl.SSLSocket) -> None:
        """Extract all data after the packet length from the socket, if needed.

        :raises ClosedSocketError: The socket is closed."""
        assert self.length.complete, "Can't extract data without complete length"
        while len(self.data) < self.length.value:
            data = sock.recv(self.length.value - len(self.data))
            if not data:
                raise ClosedSocketError("Socket closed")
            self.data.extend(data)

    def decode(self, sock: socket.socket | ssl.SSLSocket) -> MQTTPacket | None:
        """Decode a single packet straight from the socket.

        :return: None if the socket doesn't have enough data for us to decode a packet.
        :raises ClosedSocketError: The socket is closed."""
        try:
            self._extract_head(sock)
            self._extract_length(sock)
            self._extract_data(sock)
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError, WantReadError):
            # If the socket is open but doesn't have enough data for us, we need to wait for more.
            return None
        except OSError as exc:
            raise ClosedSocketError("Socket closed") from exc

        # We have a complete packet, decode it and clear the read buffer.
        packet_head = self.head
        packet_data = memoryview(self.data)
        packet_data.toreadonly()
        try:
            return decode_packet_from_parts(packet_head, packet_data)
        finally:
            packet_data.release()
            self.reset()



from typing import Any

import pytest
from pytest_mock import MockerFixture

from ohmqtt.protected import Protected, protect


class ProtectedClass(Protected):
    """A class that inherits from Protected."""
    def __init__(self, value: Any) -> None:
        super().__init__()
        self._value = value

    @protect
    def get_value(self) -> Any:
        return self.value

    @protect
    def set_value(self, value: Any) -> None:
        self._value = value

    @property
    @protect
    def value(self) -> Any:
        return self._value

    @value.setter
    @protect
    def value(self, value: Any) -> None:
        self._value = value



def test_protected_class() -> None:
    """Test the Protected class."""
    obj = ProtectedClass(42)

    with pytest.raises(RuntimeError):
        obj.get_value()
    with pytest.raises(RuntimeError):
        obj.set_value(42)
    with pytest.raises(RuntimeError):
        obj.value
    with pytest.raises(RuntimeError):
        obj.value = 42

    with obj as p:
        p.set_value(99)
        assert p.get_value() == 99
        p.set_value(100)
        assert p.value == 100


def test_protected_weird_lock(mocker: MockerFixture) -> None:
    lock_cls = mocker.patch("threading.Lock")
    lock = lock_cls()
    del lock._is_owned  # noqa: SLF001
    with pytest.raises(RuntimeError):
        Protected(lock)



from __future__ import annotations

import ssl
from typing import Final, TYPE_CHECKING

from .base import FSMState
from .closed import ClosedState
from .websocket_handshake_response import WebsocketHandshakeResponseState
from ..types import ConnectParams, StateData, StateEnvironment
from ..wslib import generate_nonce
from ...logger import get_logger

if TYPE_CHECKING:
    from ..fsm import FSM

logger: Final = get_logger("ohmqtt.connection.states.websocket_handshake_request")


class WebsocketHandshakeRequestState(FSMState):
    """Sending Websocket handshake request to the broker."""
    @classmethod
    def enter(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams) -> None:
        state_data.ws_nonce = generate_nonce()
        host = f"{params.address.host}:{params.address.port}" if not params.address.is_default_port() else params.address.host
        req = f"""GET {params.address.path} HTTP/1.1\r
Host: {host}\r
Upgrade: websocket\r
Connection: Upgrade\r
Sec-WebSocket-Key: {state_data.ws_nonce}\r
Sec-WebSocket-Version: 13\r
Sec-WebSocket-Protocol: mqtt\r
\r\n"""
        logger.debug("---> %s", req)
        state_data.write_buffer.clear()
        state_data.write_buffer.extend(req.encode("utf-8"))

    @classmethod
    def handle(cls, fsm: FSM, state_data: StateData, env: StateEnvironment, params: ConnectParams, max_wait: float | None) -> bool:
        if state_data.timeout.exceeded():
            logger.debug("Websocket handshake request keepalive timeout")
            fsm.change_state(ClosedState)
            return True

        try:
            num_sent = state_data.sock.send(state_data.write_buffer)
            if num_sent == 0:
                logger.error("Websocket handshake request send returned 0 bytes, closing connection")
                fsm.change_state(ClosedState)
                return True
            if num_sent < len(state_data.write_buffer):
                # Not all data was sent, wait for writable again.
                logger.debug("Not all Websocket handshake request data was sent, waiting for writable again: wrote: %d", num_sent)
                del state_data.write_buffer[:num_sent]
                return False
            state_data.write_buffer.clear()
            fsm.change_state(WebsocketHandshakeResponseState)
            return True
        except (BlockingIOError, ssl.SSLWantReadError, ssl.SSLWantWriteError):
            # The write was blocked, wait for the socket to be writable.
            if max_wait is None or max_wait > 0.0:
                with fsm.selector:
                    timeout = state_data.timeout.get_timeout(max_wait)
                    fsm.selector.select(write=True, timeout=timeout)
        return False



from unittest.mock import Mock

import pytest
from pytest_mock import MockerFixture

from ohmqtt.connection.keepalive import KeepAlive, MIN_TIMEOUT


@pytest.fixture
def mock_keepalive_time(mocker: MockerFixture) -> Mock:
    """Mock the keepalive module's time method to control the passage of time."""
    return mocker.patch("ohmqtt.connection.keepalive._time")


def test_keepalive_props() -> None:
    """Test the properties of the KeepAlive class."""
    keepalive = KeepAlive()
    assert keepalive.keepalive_interval == 0
    keepalive.keepalive_interval = 10
    assert keepalive.keepalive_interval == 10
    with pytest.raises(ValueError):
        keepalive.keepalive_interval = -1
    with pytest.raises(ValueError):
        keepalive.keepalive_interval = 65536


def test_keepalive_slots() -> None:
    """Ensure slots match attributes exactly."""
    keepalive = KeepAlive()
    assert not hasattr(keepalive, "__dict__")
    assert all(hasattr(keepalive, attr) for attr in keepalive.__slots__), \
        [attr for attr in keepalive.__slots__ if not hasattr(keepalive, attr)]


def test_keepalive_no_interval(mock_keepalive_time: Mock) -> None:
    """When interval is 0, keepalive should be inert."""
    keepalive = KeepAlive()

    mock_keepalive_time.return_value = 0.0
    keepalive.mark_init()
    assert keepalive.get_next_timeout() is None

    mock_keepalive_time.return_value = 1e9
    assert not keepalive.should_send_ping()
    assert not keepalive.should_close()
    assert keepalive.get_next_timeout() is None


def test_keepalive_dark_forest(mock_keepalive_time: Mock) -> None:
    """When interval is set and no data from server, close after a ping."""
    keepalive = KeepAlive()

    mock_keepalive_time.return_value = 0.0
    keepalive.mark_init()
    keepalive.keepalive_interval = 10
    assert keepalive.get_next_timeout() == 10.0

    mock_keepalive_time.return_value = 5.0
    assert not keepalive.should_send_ping()
    assert not keepalive.should_close()
    assert keepalive.get_next_timeout() == 5.0

    mock_keepalive_time.return_value = 10.0
    assert keepalive.should_send_ping()
    assert not keepalive.should_close()
    keepalive.mark_ping()
    assert keepalive.get_next_timeout() == 10.0

    mock_keepalive_time.return_value = 20.0
    assert not keepalive.should_send_ping()
    assert keepalive.should_close()
    assert keepalive.get_next_timeout() == MIN_TIMEOUT


def test_keepalive_send(mock_keepalive_time: Mock) -> None:
    """When interval is set and data is sent, no pings should be sent."""
    keepalive = KeepAlive()

    mock_keepalive_time.return_value = 0.0
    keepalive.mark_init()
    keepalive.keepalive_interval = 10

    for n in range(0, 60, 5):
        mock_keepalive_time.return_value = float(n)
        keepalive.mark_send()
        assert not keepalive.should_send_ping()
        assert not keepalive.should_close()
        assert keepalive.get_next_timeout() == 10.0


def test_keepalive_pingpong(mock_keepalive_time: Mock) -> None:
    """When interval is set now data is sent, exchange pings and pongs."""
    keepalive = KeepAlive()

    mock_keepalive_time.return_value = 0.0
    keepalive.mark_init()
    keepalive.keepalive_interval = 10

    for n in range(10, 60, 10):
        mock_keepalive_time.return_value = float(n)
        assert keepalive.get_next_timeout() == MIN_TIMEOUT
        assert keepalive.should_send_ping()
        assert not keepalive.should_close()
        keepalive.mark_ping()
        keepalive.mark_pong()
        assert keepalive.get_next_timeout() == 10.0


def test_keepalive_max_wait(mock_keepalive_time: Mock) -> None:
    """Test setting a max_wait."""
    keepalive = KeepAlive()

    assert keepalive.get_next_timeout(max_wait=1.1) == 1.1

    mock_keepalive_time.return_value = 0.0
    keepalive.mark_init()
    keepalive.keepalive_interval = 10
    assert keepalive.get_next_timeout(max_wait=2.2) == 2.2



"""SUBSCRIBE, SUBACK, UNSUBSCRIBE, and UNSUBACK packets."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Final, Sequence

from .base import MQTTPacket
from ..error import MQTTError
from ..mqtt_spec import MQTTPacketType, MQTTReasonCode
from ..property import (
    MQTTSubscribeProps,
    MQTTSubAckProps,
    MQTTUnsubscribeProps,
    MQTTUnsubAckProps,
)
from ..serialization import (
    encode_uint8,
    encode_uint16,
    encode_string,
    encode_varint,
    decode_uint8,
    decode_uint16,
    decode_string,
)

HEAD_SUBSCRIBE: Final = (MQTTPacketType.SUBSCRIBE << 4) + 0x02
HEAD_SUBACK: Final = MQTTPacketType.SUBACK << 4
HEAD_UNSUBSCRIBE: Final = (MQTTPacketType.UNSUBSCRIBE << 4) + 0x02
HEAD_UNSUBACK: Final = MQTTPacketType.UNSUBACK << 4


@dataclass(match_args=True, slots=True)
class MQTTSubscribePacket(MQTTPacket):
    packet_type = MQTTPacketType.SUBSCRIBE
    props_type = MQTTSubscribeProps
    topics: Sequence[tuple[str, int]] = field(default_factory=tuple)
    packet_id: int = 0
    properties: MQTTSubscribeProps = field(default_factory=MQTTSubscribeProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"topics={self.topics}",
            f"properties={self.properties}",
        ]
        return f"SUBSCRIBE[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        encoded = bytearray()
        encoded.append(HEAD_SUBSCRIBE)
        data = encode_uint16(self.packet_id)
        data += self.properties.encode()
        for topic, subscribe_opts in self.topics:
            data += encode_string(topic) + encode_uint8(subscribe_opts)
        encoded.extend(encode_varint(len(data)))
        encoded.extend(data)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTSubscribePacket:
        if flags != 0x02:
            raise MQTTError(f"Invalid flags, expected 0x02 but got {flags}", MQTTReasonCode.MalformedPacket)
        offset = 0
        packet_id, packet_id_length = decode_uint16(data[offset:])
        offset += packet_id_length
        props, props_length = MQTTSubscribeProps.decode(data[offset:])
        offset += props_length
        topics = []
        while offset < len(data):
            topic, topic_length = decode_string(data[offset:])
            offset += topic_length
            subscribe_opts, subscribe_opts_length = decode_uint8(data[offset:])
            offset += subscribe_opts_length
            topics.append((topic, subscribe_opts))
        if not topics:
            raise MQTTError("No topics in SUBSCRIBE packet", MQTTReasonCode.ProtocolError)
        return MQTTSubscribePacket(topics, packet_id, properties=props)


@dataclass(match_args=True, slots=True)
class MQTTSubAckPacket(MQTTPacket):
    packet_type = MQTTPacketType.SUBACK
    props_type = MQTTSubAckProps
    packet_id: int
    reason_codes: Sequence[MQTTReasonCode] = field(default_factory=tuple)
    properties: MQTTSubAckProps = field(default_factory=MQTTSubAckProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_codes={[hex(c) for c in self.reason_codes]}",
            f"properties={self.properties}",
        ]
        return f"SUBACK[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        encoded = bytearray()
        encoded.append(HEAD_SUBACK)
        length = 2 + len(self.reason_codes)
        props = self.properties.encode()
        length += len(props)
        encoded.extend(encode_varint(length))
        encoded.extend(encode_uint16(self.packet_id))
        encoded.extend(props)
        for reason_code in self.reason_codes:
            encoded.append(reason_code.value)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTSubAckPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        offset = 0
        packet_id, packet_id_length = decode_uint16(data[offset:])
        offset += packet_id_length
        props, props_length = MQTTSubAckProps.decode(data[offset:])
        offset += props_length
        reason_codes = [MQTTReasonCode(b) for b in data[offset:]]
        return MQTTSubAckPacket(packet_id, reason_codes, properties=props)


@dataclass(match_args=True, slots=True)
class MQTTUnsubscribePacket(MQTTPacket):
    packet_type = MQTTPacketType.UNSUBSCRIBE
    props_type = MQTTUnsubscribeProps
    topics: Sequence[str] = field(default_factory=tuple)
    packet_id: int = 0
    properties: MQTTUnsubscribeProps = field(default_factory=MQTTUnsubscribeProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"topics={self.topics}",
            f"properties={self.properties}",
        ]
        return f"UNSUBSCRIBE[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        encoded = bytearray()
        encoded.append(HEAD_UNSUBSCRIBE)
        data = encode_uint16(self.packet_id) + self.properties.encode()
        for topic in self.topics:
            data += encode_string(topic)
        encoded.extend(encode_varint(len(data)))
        encoded.extend(data)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTUnsubscribePacket:
        if flags != 0x02:
            raise MQTTError(f"Invalid flags, expected 0x02 but got {flags}", MQTTReasonCode.MalformedPacket)
        offset = 0
        packet_id, packet_id_length = decode_uint16(data[offset:])
        offset += packet_id_length
        props, props_length = MQTTUnsubscribeProps.decode(data[offset:])
        offset += props_length
        topics = []
        while offset < len(data):
            topic, topic_length = decode_string(data[offset:])
            offset += topic_length
            topics.append(topic)
        if not topics:
            raise MQTTError("No topics in UNSUBSCRIBE packet", MQTTReasonCode.ProtocolError)
        return MQTTUnsubscribePacket(topics, packet_id, properties=props)


@dataclass(match_args=True, slots=True)
class MQTTUnsubAckPacket(MQTTPacket):
    packet_type = MQTTPacketType.UNSUBACK
    props_type = MQTTUnsubAckProps
    packet_id: int
    reason_codes: Sequence[MQTTReasonCode] = field(default_factory=tuple)
    properties: MQTTUnsubAckProps = field(default_factory=MQTTUnsubAckProps)

    def __str__(self) -> str:
        attrs = [
            f"packet_id={self.packet_id}",
            f"reason_codes={[hex(c) for c in self.reason_codes]}",
            f"properties={self.properties}",
        ]
        return f"UNSUBACK[{', '.join(attrs)}]"

    def encode(self) -> bytes:
        encoded = bytearray()
        encoded.append(HEAD_UNSUBACK)
        data = encode_uint16(self.packet_id) + self.properties.encode()
        for reason_code in self.reason_codes:
            data += encode_uint8(reason_code.value)
        encoded.extend(encode_varint(len(data)))
        encoded.extend(data)
        return bytes(encoded)

    @classmethod
    def decode(cls, flags: int, data: memoryview) -> MQTTUnsubAckPacket:
        if flags != 0:
            raise MQTTError(f"Invalid flags, expected 0 but got {flags}", MQTTReasonCode.MalformedPacket)
        offset = 0
        packet_id, packet_id_length = decode_uint16(data[offset:])
        offset += packet_id_length
        props, props_length = MQTTUnsubAckProps.decode(data[offset:])
        offset += props_length
        reason_codes = [MQTTReasonCode(b) for b in data[offset:]]
        return MQTTUnsubAckPacket(packet_id, reason_codes, properties=props)
